/* $Id: cgeq.c,v 1.57.2.1 2007/04/08 23:32:01 pa Exp $
 * Conjugate gradients with equality constraints.
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#include "config.h"

#include "cgeq.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif /* HAVE_FLOAT_H */

/* Before we go anywhere, we need a line search routine.
 * The following encapsulates functions of one variable for said routine.
 */
typedef FLOAt (*linefunc)(FLOAt s, void *context);

/*
 * A linefunc will return a NAN (not a number) on error.
 * In C99 the <math.h> error defines a NAN macro;  for older
 * compilers, make sure we have a definition of NAN.
 * Pray that we also have an isnan().
 */
#ifndef NAN
#ifdef HAVE_WINDOWS_H
#define isnan(x) _isnan(x)
static const union {
	struct {unsigned int u0, u1;} u01;
	float f; double d;
} private_nan_= {0xffff0000, 0xffff0000};
#ifdef USE_FLOAT
#define NAN (private_nan_.f)
#else /* !USE_FLOAT */
#define NAN (private_nan_.d)
#endif /* !USE_FLOAT */
#else /* !HAVE_WINDOWS_H */
#ifdef USE_FLOAT
#define NAN (0.0F/0.0F)
#else /* !USE_FLOAT */
#define NAN (0.0/0.0)
#endif /* !USE_FLOAT */
#endif /* !HAVE_WINDOWS_H */
#endif


/* exit flags and hard-coded parameters for the line search */
#ifdef TEST
#define CGEQ_LS_ALL (\
	CGEQ_LS_CLAMP | CGEQ_LS_SADDLEPT | CGEQ_LS_BRACKETMAX |\
	CGEQ_LS_BISECTMAX | CGEQ_LS_NONLINEAR | CGEQ_LS_ERROR |\
	CGEQ_LS_FATAL \
)
#endif
static const FLOAt nl_threshold= (FLOAt) 0.1;
static const FLOAt exit_threshold= (FLOAt) 0.01;
static const int max_bracket= 10;
static const int max_bisect= 6;

/* The line search routine.
 * Look for a zero crossing of f() and report trouble as an OR of
 * the flag values above.
 *
 * Because this is an internal routine, we can make assumptions
 * that would not be warranted in a general purpose routine.
 * Also, we will let the interface be a little clumsy
 * because that allows us to reduce the work at run-time.
 */
static unsigned int line_search(
	linefunc func, FLOAt f0,
	FLOAt sguess, FLOAt smax, FLOAt *s_final,
	int guess_is_accurate,
	void *context
)
{
	/* The search always starts at s==0 and proceeds to positive s.
	 * The caller passes us the initial function value f0 == func(0)
	 * so we can save a function call.  We can assume that f0 is positive.
	 *
	 * We assume that sguess is large enough that func(sguess)-func(0)
	 * is not buried in roundoff noise, but not so large that func() is
	 * pathological in the interval [0,sguess].  Basically, we just
	 * need an estimate within a couple orders of magnitude so we can
	 * start the search.
	 *
	 * Sometimes the caller can do much better and give us an
	 * asymptotically correct sguess.  If so, he should pass an
	 * argument guess_is_accurate of 1 and we will let him know if
	 * he was right.
	 *
	 * The limit smax is advice by the caller that the function is
	 * reasonably well-behaved in the range [0,smax], and we in turn
	 * promise not to venture beyond, and also to let him know if we
	 * had to clamp the search at smax.  It is permissible to have
	 * sguess > smax, it just forces us to truncate.
	 *
	 * We try to reduce abs(func(s)) to exit_threshold * func(0) using
	 * as few function calls as we can.  If we run into trouble,
	 * we'll try to return an s>0 such that func(s) is still positive.
	 * Reason for that:  the func() is really the negative directional
	 * of an energy, func(s)= -dE(s)/ds, and we are doing a line
	 * search to minimize E(s).  If s is to the left of the
	 * zero crossing, i.e. func(s) is positive in the interval [0,s],
	 * then E(s) is guaranteed to be < E(0) and the algorithm that
	 * is calling us can make progress minimizing E.  If s is to the
	 * right of the zero crossing there is a chance that E(s) passed
	 * the minimum and increased back above the value at E(0), so we
	 * try to avoid that.
	 *
	 * Finally, we make one weird promise:  the value of s_final
	 * that we return is also the last value on which we called
	 * func().  The reasons will become clear later.
	 */
	unsigned status= CGEQ_LS_OK;
	int i;
	FLOAt sa= 0, fa= f0, sb,fb, s,f, tmp;
	int clamped;
	const FLOAt goal= f0*exit_threshold;

	/* First function evaluation.  If the function is nice and the
	 * caller's guess is accurate, this is all we'll ever need.
	 */
	s= sguess;
	if(s>smax) {s= smax; status|= CGEQ_LS_CLAMP;}
	f= func(s, context);
	if(isnan(f)) {
		status|= CGEQ_LS_ERROR | CGEQ_LS_FATAL;
		return status;
	}

	if(! guess_is_accurate) {
		/* We expect to need a better one.  Use regula falsi. */
		if(f >= f0) {
			/* Regula falsi isn't going to work, so give up.
			 * Energy surface has wrong curvature, so flag it.
			 */
			status|= CGEQ_LS_SADDLEPT;
		} else {
			s*= f0/(f0-f);
			if(s>smax) {
				s= smax;
				status|= CGEQ_LS_CLAMP;
			}
			f= func(s, context);	/* second function eval */
			if(isnan(f)) {
				status|= CGEQ_LS_ERROR | CGEQ_LS_FATAL;
				return status;
			}
		}
	}

	/* So now we have a point (s,f) that should be good
	 * if the function is nice.  Exit if within goal.
	 * If not, see how bad it was and maybe flag the function as
	 * nonlinear.
	 */
	tmp= FABS(f);
	if(tmp < goal) {
		*s_final= s;
		return status;
	} else if(tmp > f0*nl_threshold) {
		status|= CGEQ_LS_NONLINEAR;
	}

	/* Okay, if we get here we have the initial point (sa, fa)
	 * and one new point (s, f) that didn't meet convergence.
	 * Check for a sign change and, if not, bracket.
	 */
	sb= s; fb= f;
	s= -1;
	clamped= 0;
	for(i= 0; i<max_bracket; i++) {
		if(fb<0) break;
		if(clamped) {status|= CGEQ_LS_CLAMP; break;}
		s= sb; f= fb;
		/* Normally we would just double the interval length.
		 * However, it is possible that fb, while still positive,
		 * was close to zero (although not so close as to meet the
		 * exit threshold, or we would have returned already.)
		 * In this case it is aesthetically repugnant to double
		 * the interval when we were so close.  Therefore:
		 * ON THE FIRST STEP ONLY, if fb was within nl_threshold
		 * of zero, stretch the interval to try to hit f=
		 * -nl_threshold*fa.
		 */
		if( i>0 || (status & CGEQ_LS_NONLINEAR) ) sb*= 2;
		else {
			sb*= f0*(1+nl_threshold)/(f0-fb);
		}
		if(sb>smax) {sb= smax; clamped= 1;}
		fb= func(sb, context);
		if(isnan(fb)) {
			/* Blew up trying to bracket, so restore the last
			 * point.  This is the answer we will return, so
			 * re-evaluate the function because of our weird
			 * promise about always returning the last evaluated
			 * function point.
			 */
			status|= CGEQ_LS_ERROR;
			sb= s;
			fb= func(sb, context);
			clamped= 1;
			break;
		}
	}

	/* Do we have a bracket ? */
	if(fb > 0) {
		/* No.  All indications are that func(s) is positive on
		 * the interval 0 < s < smax, so returning sb==smax is the
		 * best we can do.
		 */
		status|= CGEQ_LS_BRACKETMAX;
		*s_final= sb;
		return status;
	}

	/* If we get here, we have a bracket (sa,fa) and (sb,fb)
	 * with sa<sb, fa>0, fb<0.  We may also have an intermediate
	 * point (s,f) with sa<s<sb; if not, we'll get one right away.
	 * Refine the bracketing interval by bisection.
	 *
	 * The loop invariants are:
	 *   sa < sb,  func(sa)>0, func(sb)<0 .
	 *
	 * Also,
	 *   sa < s < sb
	 * except when the evaluation of func(s) returns an error,
	 * in which case we set s to -1 
	 */
	if(s<0) {
		s= (fa*sb-fb*sa)/(fa-fb); f= func(s, context);
		if(isnan(f)) {
			/* Can't even set up, nothing we can do. */
			status|= CGEQ_LS_ERROR | CGEQ_LS_FATAL;
			return status;
		}
	}
	for(i= 0; i<max_bisect; i++) {
	if(FABS(f)<goal) {
		*s_final= s;
		return status;
	}
	if(f>0) {sa= s; fa= f;}
	else	{sb= s; fb= f;}
	s= (fa*sb-fb*sa)/(fa-fb);	/* linear interpolation */
	f= func(s, context);
	if(isnan(f)) {
		/* very bad, but maybe we can still return something useful */
		s= -1;
		status|= CGEQ_LS_ERROR; break;
	}
	}

	/* If we get here, bisection didn't work.
	 * If the left limit of the interval is > 0, we can return sa as a
	 * mediocre but safe answer.  (If sa is still ==0 that would mean
	 * no progress and we can't do that.)
	 *
	 * This is the second case where we try to return a point other than
	 * the last function evaluation point, so we have to recompute
	 * func(sa).
	 */
	status|= CGEQ_LS_BISECTMAX;
	if(sa>0) {
		*s_final= sa;
		fa= func(sa, context);
		if(isnan(fa)) {
			status|= (CGEQ_LS_ERROR | CGEQ_LS_FATAL);
			/* can only happen if the function is buggy, since
			 * this was a repeat of an earlier call that succeeded.
			 */
		}
		return status;
	}

	/* If we get here, what sometimes happens is that func() decreases
	 * rapidly to a small negative value and hangs there.  We found a
	 * bracket, but our bisection strategy by linear interpolation
	 * just kept shaving the right endpoint and never moved the
	 * left endpoint.  It is not safe to return the right endpoint
	 * so we'll go through one more bisection cycle,
	 * this time with a zero-order algorithm.
	 *
	 * It's also possible that func() returned an error (NAN) in the
	 * middle of out bisection loop.  This is very bad, but we can try
	 * again with the new bisection strategy and see what happens.
	 */
	/* --FIXME-- duplicate code.  Can we consolidate ? */
	if(s<0) {
		s= (sa+sb)/2; f= func(s, context);
		if(isnan(f)) {
			status|= CGEQ_LS_ERROR | CGEQ_LS_FATAL;
			return status;
		}
	}
	for(i= 0; i<max_bisect; i++) {
		if(FABS(f)<goal) {
			*s_final= s;
			return status;
		}
		if(f>0) {sa= s; fa= f;}
		else	{sb= s; fb= f;}
		s= (sa+sb)/2;
		f= func(s, context);
		if(isnan(f)) {
			status|= CGEQ_LS_ERROR;
			s= -1;
			break;
		}
	}

	/* Attempt partial recovery from failed bisection again. */
	if(sa>0) {
		*s_final= sa;
		fa= func(sa, context);
		if(isnan(fa)) {
			status|= (CGEQ_LS_ERROR | CGEQ_LS_FATAL);
		}
		return status;
	}

	/* Now we give up in despair. */
	*s_final= s;
	status|= CGEQ_LS_FATAL;
	return status;
} /* line_search() */


#ifdef TEST

#include "test-rig.c"
/* see comments in the test-rig file */
#include "cgeq_simplehook.c"


/* Test functions for the line search.
 * All take a (FILE *) as their context and, if not NULL,
 * use it to log their argument and return value.
 */
/* linear */
static FLOAt f0(FLOAt s, void *outfile) {
	FLOAt f= 1-s;
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* quadratic */
static FLOAt f1(FLOAt s, void *outfile) {
	FLOAt f= (1-s)*(2-s);
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* exponential */
static FLOAt f2(FLOAt s, void *outfile) {
	FLOAt f= exp(-s) - .36787944117144232159;
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* sneaky, (1-s)**64-epsilon */
static FLOAt f3(FLOAt s, void *outfile) {
	FLOAt f= 1-s;
	if(f<0) f= 0;
	f*= f; f*= f; f*= f; f*= f; f*= f; f*= f;
	f-= 0.02;
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* a step */
static FLOAt f4(FLOAt s, void *outfile) {
	FLOAt f= 1;
	static const FLOAt breakpt= .31830988618379067154;
	if(s>breakpt) f= -1;
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* a quadratic with the wrong slope initially */
static FLOAt f5(FLOAt s, void *outfile) {
	FLOAt f= (1+2*s)*(1-s);
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* a decidedly noncooperative function */
static FLOAt f6(FLOAt s, void *outfile) {
	if(s>0) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	if(outfile) fprintf(outfile, "% f  % f\n", s, 1.0);
	return 1;
}

/* a somewhat touchy function, well-behaved but defined only in [0,1[  */
static FLOAt f7(FLOAt s, void *outfile) {
	FLOAt f;
	if(s>=1) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	f= 1 + s*(log(1-s));
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* a hole where the zero crossing should be */
static FLOAt f8(FLOAt s, void *outfile) {
	FLOAt f= 1-s;
	if(.8 < s && s < 1.2) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* same, but with mild nonlinearity */
static FLOAt f9(FLOAt s, void *outfile) {
	FLOAt f= (1-s)*(2-s);
	if(.8 < s && s < 1.2) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* the sneaky function with a hole */
static FLOAt fa(FLOAt s, void *outfile) {
	FLOAt f= 1-s;
	if(f<0) f= 0;
	f*= f; f*= f; f*= f; f*= f; f*= f; f*= f;
	f-= 0.02;
	if(.05 < s && s < .10) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}

/* a step with a hole */
static FLOAt fb(FLOAt s, void *outfile) {
	FLOAt f= 1;
	static const FLOAt breakpt= .31830988618379067154;
	if(s>breakpt) f= -1;
	if(breakpt-0.05 < s && s < breakpt+0.05) {
		if(outfile) fprintf(outfile, "# % f  NAN\n", s);
		return NAN;
	}
	if(outfile) fprintf(outfile, "% f  % f\n", s, f);
	return f;
}


/* Test battery for the line search */
struct {
	linefunc f;		/* what arguments to pass */
	FLOAt sguess, smax;
	int accurate;
	unsigned int mask;	/* what bits to check in the result... */
	unsigned int flags;	/* ...and what we expect them to be */
} line_tests[]= {
	{f0,    1, 10,  1, CGEQ_LS_ALL, CGEQ_LS_OK},
	{f0,   .1, 10,  0, CGEQ_LS_ALL, CGEQ_LS_OK},
	{f0,   .1, 10,  1, CGEQ_LS_ALL, CGEQ_LS_NONLINEAR},
	{f0,   .1, .2,  0, CGEQ_LS_CLAMP, CGEQ_LS_CLAMP},
	{f1,   .1, 10,  0, CGEQ_LS_ALL, CGEQ_LS_NONLINEAR},
	{f1, 5e-4, 10,  1, CGEQ_LS_ALL, CGEQ_LS_NONLINEAR | CGEQ_LS_BRACKETMAX},
	{f2,    1, 10,  1, CGEQ_LS_ALL, CGEQ_LS_OK},
	{f3,    1,  2,  1, CGEQ_LS_ALL, CGEQ_LS_BISECTMAX},
	{f3,  100,100,  1, CGEQ_LS_FATAL, CGEQ_LS_FATAL},
	/* yes, this one fails */
	{f4,    1, 10,  0, CGEQ_LS_ALL,
		CGEQ_LS_BISECTMAX | CGEQ_LS_NONLINEAR
	},
	{f5,  .05, 2,   0, CGEQ_LS_ALL, CGEQ_LS_NONLINEAR | CGEQ_LS_SADDLEPT},
	{f6,    1, 10,  1, CGEQ_LS_ALL, CGEQ_LS_ERROR | CGEQ_LS_FATAL},
	{f6,    1, 10,  1, CGEQ_LS_ALL, CGEQ_LS_ERROR | CGEQ_LS_FATAL},
	{f7,    2, 10,  0, CGEQ_LS_ALL, CGEQ_LS_ERROR | CGEQ_LS_FATAL},
	{f7,   .3,.99,  0, CGEQ_LS_ERROR, 0},
	{f7,   .3,  2,  0, CGEQ_LS_ERROR, CGEQ_LS_ERROR},
	{f7,   .3,  2,  1, CGEQ_LS_ERROR | CGEQ_LS_FATAL, CGEQ_LS_ERROR},
	{f8,  1.5, 10,  0, CGEQ_LS_ALL, CGEQ_LS_ERROR | CGEQ_LS_FATAL},
	{f8,  1.5, 10,  1, CGEQ_LS_ALL,
		CGEQ_LS_ERROR | CGEQ_LS_FATAL | CGEQ_LS_NONLINEAR
	},
	{f9,   .3, 10,  0, CGEQ_LS_ERROR |
		CGEQ_LS_BISECTMAX, CGEQ_LS_ERROR | CGEQ_LS_BISECTMAX
	},
	{fa,    1, 10,  0, CGEQ_LS_ALL,
		CGEQ_LS_ERROR | CGEQ_LS_BISECTMAX | CGEQ_LS_FATAL
	},
	{fa,    5,100,  1, CGEQ_LS_ALL,
		CGEQ_LS_ERROR | CGEQ_LS_BISECTMAX | CGEQ_LS_FATAL
	},
	{fa,  .20, 10,  0, CGEQ_LS_ALL,
		CGEQ_LS_FATAL | CGEQ_LS_ERROR | CGEQ_LS_BISECTMAX
	},
	{fb,    1, 10,  0, CGEQ_LS_ALL,
		CGEQ_LS_ERROR | CGEQ_LS_BISECTMAX | CGEQ_LS_NONLINEAR
	},
};
static const int n_linetests= (sizeof line_tests)/(sizeof line_tests[0]);

#include <stdio.h>
int test_line_search(int argc, char *argv[]) {
	int isbad= 0;
	unsigned status;
	linefunc f;
	FLOAt s0, smax, s;
	FLOAt f0, goal, ffinal;
	int acc;
	unsigned int mask, expect;
	int i;
	FILE *out= NULL, *gnuplot= NULL;
	char *outname= NULL;

	if(verbosity>0) {
		out= openlogfile("ls", &outname);
		gnuplot= openlogfile("ls-gp", 0);
		if(gnuplot) fputs(
			"# Run this file through gnuplot "
			"to graph the line searches\n#\n"
			"set pointsize 2\n"
			"set zeroaxis xy\n\n"
			, gnuplot
		);
	}
	for(i= 0; i<n_linetests; i++) {
		if(out) fputs("\n\n", out);
		f=      line_tests[i].f;
		s0=     line_tests[i].sguess;
		smax=   line_tests[i].smax;
		acc=    line_tests[i].accurate;
		mask=   line_tests[i].mask;
		expect= line_tests[i].flags;
		f0= f(0, out);
		goal= f0*exit_threshold;
		status= line_search(f, f0, s0, smax, &s, acc, out);
		ffinal= f(s, NULL);
		if(out) fprintf(out, "# status= %#x\n", status);
		if((status&mask) != expect) {
			isbad++;
			if(out) fprintf(out,
				"# expected %#x after masking with %#x\n",
				expect, mask
			);
			printf(
				"line test % 2d: got %#x, "
				"expected %#x after masking %#x\n",
				i, status, expect, mask
			);
		}
		if(gnuplot) fprintf(gnuplot,
			"plot '%s' index %d with points 3\n"
			"pause -1\n\n"
			, outname, i
		);
		if(0 == (
			status & ~(CGEQ_LS_NONLINEAR | CGEQ_LS_SADDLEPT)
		) ) {
			/* check final function value is in bounds */
			if(FABS(ffinal) > goal) {
				isbad++;
				printf(
					"line test %d: |%e| > %e\n",
					i, ffinal, f0*exit_threshold
				);
			}
		}
	}

	if(out) fclose(out);
	if(gnuplot) fclose(gnuplot);
	if(outname) free(outname);

	if(isbad) puts("Test failed.");
	return isbad;
} /* test_line_search() */

#endif /* TEST */


/* Okay, now we set up to call the line search from a
 * conjugate gradient driver.  The function we want to compute is the
 * dot product of p[] with the coprojected gradient g[] evaluated
 * at N(x[]+s*p[]).  The "context" argument that we'll tunnel through
 * line_search() has to have enough information to compute all that,
 * which means that we have to wrap our own context around the user
 * context.
 */
struct ls_context {
	const int n;		/* system size */
	const FLOAt *x;		/* x[n], initial state */
	const FLOAt *p;		/* p[n], search direction */
	FLOAt *tmp;		/* tmp[n], to hold N(x[]+s*p[]) */
	FLOAt *g;		/* g[n], to store the gradient */
	DECORATE(cgeq_grad_proc) gradient;
	DECORATE(cgeq_normalize_proc) normalize;
	DECORATE(cgeq_coproject_proc) coproject;
	void *user_context;	/* their own baggage */
};

/* Now we can write the function of one variable that cgeq() will pass to
 * line_search().
 */
static FLOAt slope(FLOAt s, void *tunnelled) {
	const struct ls_context *c= tunnelled;
	int i; const int n= c->n;
	int user_status;
	const FLOAt *x= c->x, *p= c->p;
	FLOAt *tmp= c->tmp, *g= c->g;
	FLOAt sum= 0;

	for(i= 0; i<n; i++) tmp[i]= x[i] + s*p[i];

	if(c->normalize) {
		user_status= c->normalize(n, tmp, c->user_context);
		if(user_status<0) return NAN;
	}

	c->gradient(n, tmp, g, c->user_context);

	if(c->coproject) {
		c->coproject(n, tmp, g, c->user_context);
	}

	for(i= 0; i<n; i++) sum+= p[i]*g[i];
	return sum;
	/*
	 * Side effect:
	 * Recall that the step length returned by line_search() is
	 * also the last point where it evaluated its function.
	 * THEREFORE, a successful line search on slope() leaves the final
	 * state in c->tmp[] and the final coprojected gradient in c->g[].
	 */
} /* slope() */


int DECORATE(cgeq)(
	int n, FLOAt x[], FLOAt scratch[],
	FLOAt r2goal, FLOAt *r2, int max_iters,
	struct DECORATE(cgeq_spec) *problem,
	void *context,
	int (*logging_hook)(int opcode, void *hook_context, ...),
	void *hook_context
) {
	struct ls_context c= {
		n, x, 0, 0, 0,		/* .n, .x, .p, .tmp, .g */
		problem->gradient,	/* .gradient */
		problem->normalize,	/* .normalize */
		problem->coproject,	/* .coproject */
		context			/* .user_context */
	};
	FLOAt *work= scratch;
	FLOAt *g, *p, *z, *xtmp;
	const int get_own_space= (0==work);
	int status= CGEQ_OK;
	int i, iter;
	int u_status, ls_status;
	int hook_status= 0;
	int noisiness;
	static const int max_noisiness= 20;
	int roll_count;			/* iterations without restart */
	static const int minroll= 10;
	FLOAt g1Wg1, g1Wg0, g0Wg0;
	FLOAt s0, smax, s, beta;
	int accurate_guess, must_restart;

	/* When the algorithm doesn't go smoothly we increment the
	 * "noisiness" variable when we think the problems are caused
	 * by roundoff;  we reset to zero when we think the problems are
	 * just due to nonlinearity in the problem (because if we have
	 * enough numerical resolution to detect nonlinearity it's
	 * unlikely that we are also seeing roundoff).  We bail out if
	 * noisiness exceeds a threshold.
	 *
	 * The noise detection is skipped if the "noisiness" variable
	 * is negative.  To reduce the risk of false positives, we enable
	 * it only after we've seen minroll consecutive perfect line searches. 
	 */
	noisiness= -1;

	if(get_own_space) {
		work= malloc(4*n*sizeof(*work));
	}
	if(0==work) {status= CGEQ_ERROR; goto out;}
	g= work; p= g+n; z= p+n; xtmp= z+n;
	c.x= x; c.g= g; c.p= p; c.tmp= xtmp;

	/* Let's enforce the constraints in case they
	 * forgot.  This will either fix their mistake or fail,
	 * in which case we return a fatal error.
	 */
	if(problem->normalize) {
		if( 0 > problem->normalize(n, x, context) ) {
			status= CGEQ_CONSTRAINT_ERROR;
			g1Wg1= -1;	/* bogus squared residual */
			/* --FIXME--
			 * return FLOAT_MAX or DOUBLE_MAX or even
			 * an IEEE infinity, but use autoconf to
			 * do it right.
			 */
			goto out;
		}
	}

	problem->gradient(n, x, g, context);
	if(problem->coproject) {
		problem->coproject(n, x, g, context);
	}
	roll_count= 0;
	ls_status=  0;
	g1Wg1= g1Wg0= g0Wg0= 0;
	s= 0;
	/* Main loop.
	 * g[] is always the coprojected negative gradient at x[].
	 * On entry, this is true because we just set that up above.
	 * On later passes, this is true as a side-effect of the
	 * call to line_search() on our function slope().
	 */
	for(iter= 0; iter<max_iters; iter++) {

		/* z[] <-- direction of steepest descent */
		if(problem->precond) {
			problem->precond(n, x, g, z, context);
		}
		else memcpy(z, g, n*sizeof(*z));
		/* we don't tangent-project z[], instead we wait until
		 * it has been accumulated into p[]
		 */

		/* compute residual and exit if goal reached */
		for(i=0, g1Wg1= 0; i<n; i++) g1Wg1+= z[i]*g[i];
		if(logging_hook) {
			hook_status= logging_hook(
				CGEQ_LOG_NEWPOINT, hook_context,
				iter, g1Wg1, ls_status, s, 0.0,
				x, g, z, p
			);
		}
		if(CGEQ_LOG_FORCE_EXIT == hook_status) {
			status= CGEQ_USER_END; goto out;
		}
		if(g1Wg1 <= r2goal) {
			status= CGEQ_OK; goto out;
		}

		/* exit if we suspect roundoff noise */
		if(noisiness > max_noisiness) {
			/* (but allow caller to override) */
			if(logging_hook) {
				hook_status= logging_hook(
					CGEQ_LOG_NOISY, hook_context
				);
			}
			if(CGEQ_LOG_FORCE_CONTINUE != hook_status) {
				status= CGEQ_ROUNDOFF; goto out;
			}
		}

		/* Step to new state.
		 * If the line search returns an error, we have *one* chance
		 * to retry by restarting the conjugate gradients.
		 */
		attempt_step:
		/* decide if we want to restart */
		must_restart=
			(
				ls_status &&
				(CGEQ_LOG_FORCE_NORESTART != hook_status)
			) || (
				CGEQ_LOG_FORCE_RESTART == hook_status
			)
		;
		/* let them know about the restart */
		if(must_restart) {
			hook_status= 0;
			if(logging_hook && iter>0) {
				hook_status= logging_hook(
					CGEQ_LOG_RESTART, hook_context,
					ls_status, 0.0
				);
			}
			/* exit if they tell us to */
			if(CGEQ_LOG_FORCE_EXIT == hook_status) {
				status= CGEQ_ITMAX; goto out;
			}
			/* cancel the restart if they say so ... */
			if(CGEQ_LOG_FORCE_NORESTART == hook_status) {
				must_restart= 0;
				/* ... but fail if the restart was
				 * for error recovery.
				 */
				if(ls_status & CGEQ_LS_FATAL) {
					status= CGEQ_ERROR; goto out;
				}
			}
		}
		/* always "restart" at iteration 0 */
		must_restart|= (0==iter);

		/* new search direction */
		if(must_restart) {
			memcpy(p, z, n*sizeof(*p));
			roll_count= 0;
		} else {
			beta= (g1Wg1 - g1Wg0)/g0Wg0;
			for(i= 0; i<n; i++) p[i]= z[i] + p[i]*beta;
		}
		if(problem->tproject) {
			problem->tproject(n, x, p, context);
		}

		/* Get a step size.  The step guesser should never fail, so
		* if it does we just give up.
		*/
		u_status= problem->guess_step(n, x, p, &s0, &smax, context);
		if(u_status<0) {
			status= CGEQ_ERROR; goto out;
		}
		accurate_guess= (u_status>0);

		/* do the line search and carefully check its status */
		ls_status= line_search(
			slope, g1Wg1,
			s0,smax, &s, accurate_guess,
			&c
		);
		if(ls_status & CGEQ_LS_FATAL) {
			if(0 == roll_count) {
				status= CGEQ_ERROR; goto out;
			}
			/* We need to recompute the gradient */
			problem->gradient(n, x, g, context);
			if(problem->coproject) {
				problem->coproject(n, x, g, context);
			}
			goto attempt_step;
		}
		/* not fatal, so assume the line search made progress */
		memcpy(x, xtmp, n*sizeof(x[0]));

		/* save gn*W*gn and compute gn*W*gn+1 */
		g0Wg0= g1Wg1;
		for(i= 0, g1Wg0= 0; i<n; i++) g1Wg0+= g[i]*z[i];

		/* number of consecutive perfect line searches */
		if(CGEQ_LS_OK == ls_status) roll_count++;

		/* Noisiness check.
		* Repeated bisection failures are a symptom of roundoff noise.
		* Repeated linearity failures are a weak symptom of roundoff.
		* However, most other troubles are caused by nonlinearity
		* and indicate that we are nowhere near a minimum --so we
		* reset the count.  Also reset on 3 consecutive
		* successful searches --FIXME-- magic number.
		*/
		if(noisiness<0) {
			/* maybe turn on the checks */
			if(roll_count > minroll) noisiness= 0;
		} else if(
			((CGEQ_LS_OK == ls_status) && (roll_count > 3)) ||
			(ls_status &
			(CGEQ_LS_CLAMP | CGEQ_LS_SADDLEPT | CGEQ_LS_BRACKETMAX)
			)
		) {
			noisiness= 0;
		} else if(ls_status & CGEQ_LS_BISECTMAX) {
			noisiness+= 5;
		} else if(ls_status & CGEQ_LS_NONLINEAR) {
			noisiness+= 1;
		}

	} /* for(iter) */

	status= CGEQ_ITMAX;

	out:
	if(get_own_space && work) {
		free(work);
	}
	*r2= g1Wg1;
	return status;
} /* cgeq() */


#ifdef TEST

/* Solve linear systems. */

/* Step guesser can return anything as long as it doesn't claim to be
 * accurate.
 */
static int anystep(
	int n, const FLOAt x[], const FLOAt p[],
	FLOAt *s0, FLOAt *smax,
	void *ignored
) {*s0= 1; *smax= 1234; return 0;}

/* Diagonal matrix, zero RHS, solution is zero but the algorithm
 * doesn't know that !
 */
struct lindiag {
	FLOAt *diag;
};
static void lindiag_grad(
	int n, const FLOAt x[], FLOAt g[], void *context
) {
	struct lindiag *c= context;
	int i;
	for(i= 0; i<n; i++) g[i]= -x[i]*c->diag[i];
}
#define lindiag_N 20

/* Tridiagonal system -- no context needed */
static void tridiag_grad(
	int n, const FLOAt x[], FLOAt g[], void *ignored
) {
	int i;
	for(i= 0; i<n; i++) g[i]= 0;
	for(i= 0; i<n-1; i++) {
		g[  i]-= x[i] - x[i+1];
		g[i+1]+= x[i] - x[i+1];
	}
	g[0]= g[n-1]= 0;
}


/* Test the hook control feature */
#include <stdarg.h>
static int loghook(int opcode, void *logging_context, ...)
{
	FILE *out= logging_context;
	va_list ap;
	int iter; double r2, conjug; unsigned int ls_status;
	const int restart_cycle= 10, itmax= 56;
	int command= 0;


	va_start(ap, logging_context);
	switch(opcode) {
	case CGEQ_LOG_NEWPOINT:
		iter= va_arg(ap, int);
		r2= va_arg(ap, double);
		ls_status= va_arg(ap, unsigned int);
		conjug= va_arg(ap, double);	/* step size, skip */
		conjug= va_arg(ap, double);	/* conjugacy score, keep */
		if(out) fprintf(out,
			"% 4d  % 13.5e  %#4x  % 11.3e\n",
			iter, r2, ls_status, conjug
		);
		if(1 == (iter%restart_cycle)) {
			command= CGEQ_LOG_FORCE_RESTART;
		}
		if(iter > itmax) {
			command= CGEQ_LOG_FORCE_EXIT;
		}
	break;
	case CGEQ_LOG_RESTART:
		ls_status= va_arg(ap, unsigned int);
		conjug= va_arg(ap, double);
		if(out) fprintf(out,
			"# restart %#4x  % 11.3e\n\n", ls_status, conjug
		);
	break;
	}
	va_end(ap);
	return command;
} /* loghook() */

static int linsys_test(int argc, char *argv[]) {
	int i;
	FLOAt diag[lindiag_N];
	struct lindiag context= {diag};
	FLOAt x[lindiag_N], work[4*lindiag_N];
	FILE *log= NULL, *gnuplot= NULL;
	char *logname= NULL;
	const FLOAt r2goal= 1e-10;
	FLOAt r2;
	int cg_status;
	int isbad= 0;
	struct DECORATE(cgeq_spec) problem= {
		lindiag_grad,	/* .gradient */
		0,		/* .precond */
		0,		/* .normalize */
		0,		/* .tproject */
		0,		/* .coproject */
		anystep,	/* .guess_step */
	};

	if(verbosity>0) {
		log= openlogfile("cg", &logname);
		gnuplot= openlogfile("cg-gp", 0);
		if(gnuplot) fputs(
			"# Run through gnuplot to trace convergence\n"
			"# of unconstrained conjugate gradients\n\n"
			"set logscale y\n"
			"set style data linespoints\n\n"
			, gnuplot
		);
	}

	if(log) fputs("# Diagonal system\n", log);
	if(gnuplot) fprintf(gnuplot,
		"set title 'Diagonal system, n= %d'\n"
		, lindiag_N
	);
	for(i= 0; i<lindiag_N; i++) diag[i]= badrandom();
	for(i= 0; i<lindiag_N; i++) x[i]= badrandom();
	cg_status= DECORATE(cgeq)(
		lindiag_N, x, work, r2goal, &r2, 100,
		&problem, &context,
		cgeq_simple_log_hook, log
	);
	if(cg_status) {
		printf("linear CG returned %d, expected 0\n", cg_status);
		isbad++;
	}
	if(log) fprintf(log, "# cg_status= %#x\n", cg_status);
	if(gnuplot) fprintf(gnuplot,
		"plot '%s' index 0\n"
		"pause -1\n\n"
		, logname
	);

	if(log) fputs("\n\n# Tridiagonal system\n", log);
	if(gnuplot) fprintf(gnuplot,
		"set title 'Tridiagonal system, n=%d'\n"
		, lindiag_N
	);
	problem.gradient= tridiag_grad;
	for(i= 0; i<lindiag_N; i++) x[i]= badrandom();
	for(i= 0; i<5; i++) {
		cg_status= DECORATE(cgeq)(
			lindiag_N, x, 0 /* work */, 0 /* r2goal */, &r2, 100,
			&problem,
			NULL,
			cgeq_simple_log_hook, log
		);
		if(log) fprintf(log, "# cg_status= %#x\n", cg_status);
		if(gnuplot) fprintf(gnuplot,
			"plot '%s' index 1\n"
			"pause -1\n\n"
			, logname
		);
		if(CGEQ_ITMAX != cg_status) break;
	}
	if( (cg_status != CGEQ_OK) && (cg_status != CGEQ_ROUNDOFF) ) {
		printf(
			"linear CG returned %d, expected %d or %d\n",
			cg_status, CGEQ_OK, CGEQ_ROUNDOFF
		);
		isbad++;
	}

	if(log) fputs("\n\n#Tridiagonal system with forced restarts\n", log);
	if(gnuplot) fputs(
		"set title 'Tridiagonal system with forced restarts'\n"
		, gnuplot
	);
	for(i= 0; i<lindiag_N; i++) x[i]= badrandom();
	cg_status= DECORATE(cgeq)(
		lindiag_N, x, work, 0, &r2, 12345,
		&problem, NULL,
		loghook, log
	);
	if(log) fprintf(log, "# cg_status= %#x\n", cg_status);
	if(gnuplot) fprintf(gnuplot,
		"plot '%s' index 2\n"
		"pause -1\n\n"
		, logname
	);

	if(log) fclose(log);
	if(gnuplot) fclose(gnuplot);
	if(logname) free(logname);
	return !!isbad;
} /* linsys_test() */


/* Lowest eigenvalue by Rayleigh quotient.
 * Min_x x^t A x, subject to x^t x == 1.
 * Also use this test to check the CGEQ_CONSTRAINT_ERROR bit.
 */
/* normalization */
static int rayleigh_norm(int n, FLOAt x[], void *ignored) {
	int i;
	FLOAt l2;
	for(i= 0, l2= 0; i<n; i++) l2+= x[i]*x[i];
	if(l2==0) return -1;
	l2= 1/sqrt(l2);
	for(i= 0; i<n; i++) x[i]*= l2;
	return 0;
}
/* projection, also coprojection */
static void rayleigh_proj(
	int n, const FLOAt x[], FLOAt p[], void *ignored
) {
	int i;
	FLOAt dot;
	for(i= 0, dot= 0; i<n; i++) dot+= x[i]*p[i];
	for(i= 0; i<n; i++) p[i]-= x[i]*dot;
}
/* stepsize guesser */
static int rayleigh_step(
	int n, const FLOAt x[], const FLOAt p[],
	FLOAt *s0, FLOAt *smax, void *ignored
) {
	/* use the sphere radius (==1) as a characteristic scale */
	int i;
	FLOAt pmax, tmp;
	for(i= 0, pmax= 0; i<n; i++) {
		tmp= FABS(p[i]);
		if(pmax<tmp) pmax= tmp;
	}
	if(pmax<=0) return -1;
	*s0= 0.1/pmax;
	*smax= 10/pmax;
	return 0;
}
#define N_RAYLEIGH 30
static int test_rayleigh_eigen(int argc, char *argv[]) {
	FLOAt diag[N_RAYLEIGH];
	struct lindiag context= {diag};
	FLOAt x[N_RAYLEIGH], work[4*N_RAYLEIGH];
	FILE *log= NULL;
	FILE *gnuplot= NULL;
	char *logname= NULL;
	FLOAt r2, r2goal;
	FLOAt err, max_err, tmp;
	int i, repeat, insist;
	int cg_status;
	const int n= N_RAYLEIGH;
	int isbad= 0;
	struct DECORATE(cgeq_spec) problem= {
		lindiag_grad,	/* .gradient */
		0,		/* .precond */
		rayleigh_norm,	/* .normalize */
		rayleigh_proj,	/* .tproject */
		rayleigh_proj,	/* .coproject */
		rayleigh_step,	/* .guess_step */
	};

	/* -- FIXME -- boilerplate alert, think consolidation */
	if(verbosity>0) {
		log= openlogfile("rayleigh-cg", &logname);
		gnuplot= openlogfile("rayleigh-gp", 0);
		if(gnuplot) fputs(
			"# Run through gnuplot to trace convergence of\n"
			"# nonlinear CG on a Rayleigh quotient.\n\n"
			"set logscale y\n"
			"set style data linespoints\n\n"
			, gnuplot
		);
	}

	if(log) fputs("# Lowest eigenvalue by Rayleigh-Ritz\n", log);

	/* set up eigenvalues */
	diag[0]= 1.0;
	for(i= 1; i<n; i++) diag[i]= 1.2*diag[i-1];

	/* find our machine epsilon and square it */
	{
		volatile FLOAt u, v= 1;
		for(;;) {
			u= 1+v;
			u-= 1;
			if(0==u) break;
			v/= 2;
		}
		r2goal= v*v;
	}
	/* add some slop */
	r2goal*= 50;
	max_err= (r2goal ? 10.*sqrt(r2goal) : 1e-5);

	for(repeat= 0; repeat<5; repeat++) {
		for(i= 0; i<n; i++) x[i]= 2*badrandom()-1;
		/* and we don't even bother to normalize x[] ! */
		for(insist= 0; insist<5; insist++) {
			cg_status= DECORATE(cgeq)(
				n, x, work, r2goal, &r2, 100,
				&problem,
				&context,
				cgeq_simple_log_hook, log
			);
			if(log) fprintf(
				log, "# cg_status= %#x, r2= %e\n\n",
				cg_status, r2
			);
			if(CGEQ_ITMAX != cg_status) break;
		} /* for(insist) */
		if(log) fputs("\n", log);
		if(gnuplot) fprintf(gnuplot,
			"plot '%s' index %d\n"
			"pause -1\n\n"
			, logname, repeat
		);

		/* we know the lowest eigenvector, don't we */
		err= FABS(x[0])-1;
		for(i= 1; i<n; i++) {
			tmp= FABS(x[i]);
			if(err<tmp) err= tmp;
		}
		if(err > max_err) {
			isbad++;
			printf("error %e > expectation %e\n", err, max_err);
		}
	}

	/* test behavior under unrecoverable constraint error */
	if(log>0) fputs(
		"\n\n# check failure from bad initial point...\n",
		log
	);
	for(i= 0; i<n; i++) x[i]= 0;
	cg_status= DECORATE(cgeq)(
		n, x, work, r2goal, &r2, 100,
		&problem, &context, cgeq_simple_log_hook, log
	);
	if(log) fprintf(log,
		"# got status %#x, expected  %#x\n"
		"# r2= %e\n",
		cg_status, CGEQ_CONSTRAINT_ERROR, r2
	);
	if(CGEQ_CONSTRAINT_ERROR != cg_status) {
		isbad++;
		printf(
			"failure to fail, got status %#x, "
			"expected %#x\n",
			cg_status, CGEQ_CONSTRAINT_ERROR
		);
		if(log) fputs(
			"# failure to fail !", log
		);
	}

	if(gnuplot) fclose(gnuplot);
	if(log) fclose(log);
	if(logname) free(logname);

	return !!isbad;
} /* test_rayleigh_eigen() */

/* The null-terminated array of tests. */
int (*tests[])(int argc, char *argv[])= {
	test_line_search, linsys_test, test_rayleigh_eigen,
	0
};

#endif /* TEST */
