/* Conjugate gradients with equality constraints
 * $Id: cgeq.h,v 1.24.2.1 2007/04/08 23:32:01 pa Exp $
 *
 * Copyright (C) 2007  Seagate Technology
 * See terms at the end of this file.
 */
#ifndef CGEQ_H
#define CGEQ_H

#ifdef __cplusplus
extern "C" {
#endif

/* We minimize a function of several variables, subject to equality
 * constraints.  The function to minimize and the constraints are
 * wrapped in a struct ...
 */
struct cgeq_spec;

/* ... which is then passed to the main entry point. */ 
extern int cgeq(
	int n,			/* system size */
	double x[],		/* x[n], initial state, replaced by solution */
	double scratch[],	/* scratch[4*n], or NULL */
	double residual2_goal,	/* exit criterion, squared residual */
	double *residual2_actual,	/* output, residual actually achieved */
	int max_iteration,		/* not to exceed */
	struct cgeq_spec *problem,	/* problem description */
	void *context,		/* pointer to arbitrary data */
	/*
	* The next two parameters will normally be passed as NULL.
	* Ignore them for now, but see the discussion at the bottom of
	* this file.
	*/
	int (*logging_hook)(int opcode, void *hook_context, ...),
	void *hook_context
);

/* cgeq() returns one of the following: */
#define CGEQ_OK		00	/* converged, no error */
#define CGEQ_ROUNDOFF	01	/* probably converged, can't improve */
#define CGEQ_ITMAX	02	/* iteration count exceeded, try again */
#define CGEQ_ERROR	03	/* something went wrong */

/* If you use the logging_hook argument you can cause
 * cgeq() to terminate early, with a return code of
 */
#define CGEQ_USER_END	04	/* terminated by logging hook */

/* The initial state x[] has to satisfy the constraints.
 * Separate error code for when the caller forgets and
 * we're unable to fix it.
 */
#define CGEQ_CONSTRAINT_ERROR	05	/* bad initial state */


/* And what goes into a struct cgeq_spec ?  Mostly pointers to a
 * bunch of functions ("callbacks") that cgeq() will call when it
 * needs to perform certain operations.  Some of these functions
 * return zero for success and a negative number for errors.
 * (cgeq() can do some error recovery, but in many cases it'll just
 * abort and return an error itself.)  Other callbacks must never
 * fail and just return void.
 *
 * All the callbacks will receive a copy of cgeq's "context"
 * argument.  That way, the user can parametrize the problem
 * without using global variables.
 */

/* Callback to compute negative gradients (should not fail) */
typedef void (*cgeq_grad_proc)(
	int n,			/* problem size */
	const double state[],	/* state[n], input data */
	double gradient[],	/* gradient[n], result */
	void *context		/* pointer to anything they need */
);

/* Optional preconditioning callback (should not fail). */
typedef void (*cgeq_precond_proc)(
	int n,
	const double state[],		/* state[n] */
	const double gradient[],	/* gradient[n] */
	double stepdir[],		/* stepdir[n], output */
	void *context
);

/* The next three callback types describe the constraints.
 * If the problem is unconstrained, they are unnecessary.
 *
 * Normalization routine takes an arbitrary point, possibly violating the
 * constraints, and returns the ``nearest'' feasible point.
 * There is considerable latitude in defining what ``nearest'' means.
 *
 * Projection routine takes a feasible point and an infinitesimal
 * displacement, possibly incompatible with the constraints.  It
 * projects the infinitesimal displacement on the subspace compatible
 * with the constraints.  This projection MUST BE THE INFINITESIMAL
 * VERSION OF THE NORMALIZATION OPERATION.
 *
 * Coprojection is the dual of projection.  It takes a generalized
 * force and removes components parallel to the constraint gradients.
 * IT MUST REALLY BE THE DUAL OF PROJECTION, i.e. the coprojection
 * matrix must be the transpose of the projection matrix.
 *
 * The normalization callback may fail if the starting point is too
 * far from the feasible manifold, so it returns an int (0==success,
 * negative==error).  The other two callbacks have no reason to fail
 * and just return void.
 */
typedef int (*cgeq_normalize_proc)(	/* can fail */
	int n, double state_inout[], void *context
);
typedef void (*cgeq_project_proc)(	/* should not fail */
	int n, const double state[], double delta[], void *context
);
typedef void (*cgeq_coproject_proc)(	/* should not fail */
	int n, const double state[], double force[], void *context
);

/* Stepsize initialization callback for the line searches.
 * Given a state x[] and a search direction p[], this callback must
 * return a scalar s0 such that (x[] + s0*p[]) is close to a directional
 * energy minimum (after normalization).  The estimate can be VERY crude,
 * but we need SOMETHING because we have no clue as to the scale of
 * the problem.
 *
 * We also need a maximum step size s_max that we promise not to
 * exceed in the line search.  We expect that, for 0 < s < s_max,
 * we can pass x[] + s*p[] to the normalization routine and not get an
 * error.
 *
 * Do NOT assume that p[] is normalized to unit Euclidean length or
 * other convention.  In fact the numerical length of p[] will get
 * pretty wild.  You must really examine the components of p[] and
 * scale the results s0 and smax accordingly.
 *
 * This callback returns an int even though it should never fail.
 * The normal return value is zero, but it can also return 1 to assert
 * that not only everything went well, but also that the estimate s0
 * is accurate.  (We use that information to decide when to restart
 * the conjugate gradient algorithm.)
 *
 * If the callback returns a negative number, we will treat that as a
 * fatal error and fail immediately.
 */
typedef int (*cgeq_stepsize_proc)(	/* should not fail */
	int n,
	const double state[], const double direction[],
	double *s0, double *smax,
	void *context
);


/* Now we can define the problem description struct. */
struct cgeq_spec {
	cgeq_grad_proc      gradient;	/* required */
	cgeq_precond_proc   precond;	/* may be NULL */
	cgeq_normalize_proc normalize;	/* NULL if unconstrained */
	cgeq_project_proc   tproject;	/* NULL if unconstrained */
	cgeq_project_proc  coproject;	/* NULL if unconstrained */
	cgeq_stepsize_proc guess_step;	/* required */
};


/* We also provide variants ending in "_f", for single precision */

typedef void (*cgeq_grad_proc_f)(
	int n, const float state[], float gradient[], void *context
);
typedef void (*cgeq_precond_proc_f)(
	int n, const float state[], const float gradient[], float stepdir[],
	void *context
);
typedef int (*cgeq_stepsize_proc_f)(
	int n, const float state[], const float direction[],
	float *s0, float *smax, void *context
);
typedef int (*cgeq_normalize_proc_f)(
	int n, float state_inout[], void *context
);
typedef void (*cgeq_project_proc_f)(
	int n, const float state[], float delta[], void *context
);
typedef void (*cgeq_coproject_proc_f)(
	int n, const float state[], float force[], void *context
);

struct cgeq_spec_f {
	cgeq_grad_proc_f		gradient;
	cgeq_precond_proc_f		precond;
	cgeq_normalize_proc_f	normalize;
	cgeq_project_proc_f		tproject;
	cgeq_project_proc_f		coproject;
	cgeq_stepsize_proc_f	guess_step;
};

extern int cgeq_f(
	int n, float x[], float scratch[],
	float residual2_goal, float *residual2, int max_iteration,
	struct cgeq_spec_f *problem, void *context,
	int (*logging_hook)(int opcode, void *hook_context, ...),
	void *hook_context
);


/* Now about the logging callback.
 *
 * If you pass non-null (logging_hook, hook_context) arguments,
 * your logging_hook() will be called at each iteration, allowing you
 * monitor the progress of the algorithm.
 *
 * The library supplies a simple logging callback, whose "hook_context" is
 * a stdio stream (aka FILE *) open for writing.  The callback just prints
 * one line of statistics per iteration.  The resulting file can be
 * plotted with gnuplot to display the convergence behavior.
 */
extern int cgeq_simple_log_hook(int opcode, void *stdio_stream, ...);
/*
 * A more sophisticated callback could update a graphical display
 * of the system state, for example.
 *
 * If you write your own logging callback, you should normally return
 * a value of zero.  However, you can fine-tune the conjugate gradient
 * algorithm by returning nonzero values as described below.
 *
 * The arguments to the logging callback are an integer opcode, the
 * hook_context, and a variadic sequence of arguments depending on the opcode.
 *
 * I may add opcodes in the future.  For forward compatibility, your
 * callback should handle unknown opcodes by doing nothing and returning zero.
 *
 * The opcodes are:
 */
#define CGEQ_LOG_NEWPOINT 1
#define CGEQ_LOG_RESTART  2
#define CGEQ_LOG_NOISY    3
/*
 * Opcode CGEQ_LOG_NEWPOINT means that an iteration just completed.
 * The full call is (in pseudocode)
 *
 *	value= logging_hook(CGEQ_LOG_NEWPOINT, hook_context,
 *	    int iter,			// iteration number
 *	    double r2,			// new squared residual
 *	    unsigned int flags,		// status bits from last line search
 *	    double s,			// last step size
 *	    double conjug,		// loss of conjugacy --BROKEN--
 *	    <double or float> *x,	// new state
 *	    <double or float> *g,	// new gradient (coprojected)
 *	    <double or float> *z,	// new steepest descent direction
 *	    <double or float> *p	// last search direction
 *	);
 *
 * The flags argument is a bitmask, the logical OR of
 */
#define CGEQ_LS_OK		0x0
#define CGEQ_LS_CLAMP		0x1	/* tried to exceed search range */
#define CGEQ_LS_SADDLEPT	0x2	/* nearby maximum or saddle point */
#define CGEQ_LS_BRACKETMAX	0x4	/* gave up trying to bracket a zero */
#define CGEQ_LS_BISECTMAX	0x8	/* gave up trying to refine a zero */
#define CGEQ_LS_NONLINEAR	0x10	/* function is provably nonlinear */
#define CGEQ_LS_CONJUG		0x20	/* loss of conjugacy --BROKEN-- */
#define CGEQ_LS_ERROR		0x40	/* function returned an error */
#define CGEQ_LS_FATAL		0x80	/* search failed */
/*
 * If the flags argument is nonzero, cgeq() will schedule a conjugate
 * gradient restart at the next iteration.
 *
 * NOTE:
 *    The test for loss of conjugacy is currently disabled.
 *    The CGEQ_LS_CONJUG bit is never set and the conjug argument
 *    is always a dummy zero.  Ignore for now.
 *
 * Allowed hook return values, besides zero, are
 */
#define CGEQ_LOG_FORCE_RESTART   1
#define CGEQ_LOG_FORCE_NORESTART 2
#define CGEQ_LOG_FORCE_EXIT      3
/* 
 * The first two force or prevent a restart at the next iteration.
 * The third causes cgeq() to exit immediately with return value
 * CGEQ_USER_END .
 *
 *
 * Opcode CGEQ_LOG_RESTART means that we're about to restart.
 * The complete call is:
 *
 *	value= logging_hook(
 *	    CGEQ_LOG_RESTART, hook_context,
 *	    unsigned int flags,		// from last line search
 *	    double conjug		// loss of conjugacy --BROKEN--
 *	);
 *
 * The arguments flags and conjug are as above and documents cgeq()'s
 * reason for wanting to restart (flags could be zero if the restart was
 * requested by you at the last CGEQ_LOG_NEWPOINT).  You can return
 * CGEQ_LOG_FORCE_NORESTART to override cgeq()'s decision and cancel
 * the restart (but it's probably easier to do that at the NEWPOINT
 * callback).
 *
 * Opcode CGEQ_LOG_NOISY means that cgeq() estimates that your error
 * goal is unreachable and is about to exit.  The complete call is
 *
 *     value= logging_hook(CGEQ_LOG_NOISY, hook_context); // no further args
 *
 * If you think that your error goal is perfectly reasonable and that
 * cgeq() is returning prematurely, you can disable the noise detector
 * by returning
 */
#define CGEQ_LOG_FORCE_CONTINUE 4

#ifdef __cplusplus
}
#endif

/* 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#endif /* CGEQ_H */
