/* $Id: minimu.c,v 1.34.2.1 2007/04/08 23:32:01 pa Exp $
 * Energy minimization, unconstrained micromagnetics
 * (besides the usual constraint on the length, that is)
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#ifdef TEST
#include "test-rig.c"

/* we depend on cgeq.c, so we inline it here without its testing code */
#undef TEST
#include "cgeq.c"
#include "cgeq_simplehook.c"
#define TEST
#endif /* TEST */

#include "config.h"

#include <math.h>
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif /* HAVE_FLOAT_H */
#include "minimu.h"
#include "cgeq.h"

/* STOP
 * Read the following header first
 */
#include "minimu_P.h"

/* cgeq() callbacks.
 * cgeq() calls us with n== number of variables,
 * we convert to n== number of spins and we call the user's code.
 * The information we need to forward the calls is in the struct
 * minimu_wrap pointed to by the context.
 *
 * You did read the header, right ?
 */
void DECORATE(minimu_grad_wrap)(
	int n, const FLOAt state[], FLOAt gradient[], void *context
) {
	struct DECORATE(minimu_wrap) *w= context;
	w->get_field(n/3, state, gradient, w->minimu_context);
}
void DECORATE(minimu_precond_wrap)(
	int n, const FLOAt state[], const FLOAt grad[],
	FLOAt step[], void *context
) {
	struct DECORATE(minimu_wrap) *w= context;
	w->get_steepest(n/3, state, grad, step, w->minimu_context);
}

/* Machine epsilon.
 * We need this to avoid denormals in minimu_normalize() below.
 * FIXME do this in autoconf.
 */
static FLOAt epsilon_= -1;
static void init_epsilon(void)
{
	volatile FLOAt tmp= 1;
	epsilon_= 1;
	while(tmp) {
		epsilon_/= 2;
		tmp= 1-epsilon_;
		tmp-= 1;
	}
}


/* cgeq callbacks (see cgeq.h).
 * These functions used to be static but they are now exported
 * by minimu_P.h so minimu_push.c can reuse them.  They are not part
 * of the normal user interface.
 */

/* Micromagnetic normalization */
int DECORATE(minimu_normalize)(int n, FLOAt s[], void *ignored) {
    int i;
    FLOAt sx, sy, sz, r;
    if(epsilon_<0) init_epsilon();
    for(i= 0; i<n; i+= 3) {
	sx= s[i]; sy= s[i+1]; sz= s[i+2];
	r= sx*sx + sy*sy + sz*sz;
	if(r<=0) return -1;
	r= 1/SQRT(r);
	sx*= r; sy*= r; sz*= r;
	/* Flush small components to zero.
	 * It doesn't do any harm to the accuracy because all three
	 * components can't be small at the same time.
	 * If we don't do it, sooner or later we generate denormals
	 * and pay a big performance penalty on Intel.
	 */
	if(FABS(sx)<epsilon_) sx= 0;
	if(FABS(sy)<epsilon_) sy= 0;
	if(FABS(sz)<epsilon_) sz= 0;
	s[i]= sx; s[i+1]= sy; s[i+2]= sz;
    }
    return 0;
}
/* Micromagnetic tangent and cotangent projection */
void DECORATE(minimu_project)(
	int n, const FLOAt s[], FLOAt ds[], void *ignored
) {
	int i;
	FLOAt sx,sy,sz, px,py,pz, dot;
	for(i= 0; i<n; i+= 3) {
		sx=  s[i]; sy=  s[i+1]; sz=  s[i+2];
		px= ds[i]; py= ds[i+1]; pz= ds[i+2];
		dot= sx*px + sy*py + sz*pz;
		ds[i]= px - sx*dot; ds[i+1]= py - sy*dot; ds[i+2]= pz - sz*dot;
	}
}

/* Step size guesser.
 * We know the scale of the problem very well, since all spin
 * components are in [-1 : 1] and since micromagnetic energies
 * are pretty smooth.  We make them take a step where the fastest spin
 * moves by 0.5 to 1 degree, and we clamp the maximum rotation to about
 * 35 degrees.
 *
 * This doesn't need to be accurate, so we use the Cartesian
 * components directly without bothering to calculate Euclidean norms.
 */
int DECORATE(minimu_stepsize)(
	int n, const FLOAt s[], const FLOAt ds[],
	FLOAt *s0, FLOAt *smax,
	void *ignored
) {
	int i;
	FLOAt tmp, max;
	for(max= 0, i= 0; i<n; i++) {
		tmp= FABS(ds[i]);
		if(max<tmp) max= tmp;
	}
	if(0==max) return -1;
	*s0= 0.01/max; *smax= 0.4/max;
	/* s0:   atan(.01)= 0.6 degree, atan(.01*sqrt(3))=  1 degree.
	 * smax: atan(0.4)= 22 degrees, atan(0.4*sqrt(3))= 35 degrees
	 */
	return 0;
}


/* The main routine */
int DECORATE(minimu)(
	int n,
	FLOAt s[], FLOAt scratch[],
	FLOAt rtorq2_goal, FLOAt *rtorq2, int max_iterations,
	DECORATE(minimu_field_proc) get_field,
	DECORATE(minimu_precond) get_deltaspin,
	void *context,
	int (*logging_hook)(int opcode, void *logging_context, ...),
	void *logging_context
) {
	/* stuff the user callbacks into our own struct... */
	struct DECORATE(minimu_wrap) w= {
		get_field,	/* .get_field */
		get_deltaspin,	/* .get_steepest */
		context		/* .context */
	};
	/* ... stuff our own callbacks in cgeq's struct... */
	struct DECORATE(cgeq_spec) problem= {
		DECORATE(minimu_grad_wrap),	/* .gradient */
		(get_deltaspin ?
			DECORATE(minimu_precond_wrap) :
			0
		),				/* .precond */
		DECORATE(minimu_normalize),	/* .normalize */
		DECORATE(minimu_project),	/* .tproject */
		DECORATE(minimu_project),	/* .coproject */
		DECORATE(minimu_stepsize)	/* .guess_step */
	};
	/* ... and go. */
	return DECORATE(cgeq)(
		3*n, s, scratch, rtorq2_goal, rtorq2, max_iterations,
		&problem, &w,
		logging_hook, logging_context
	);
} /* minimu() */

/* but we'd better check for skew between the return codes */
#if MINIMU_OK != CGEQ_OK
#error "fix definition of MINIMU_OK in minimu.h"
#endif
#if MINIMU_ROUNDOFF != CGEQ_ROUNDOFF
#error "fix definition of MINIMU_ROUNDOFF in minimu.h"
#endif
#if MINIMU_ITMAX != CGEQ_ITMAX
#error "fix definition of MINIMU_ITMAX in minimu.h"
#endif
#if MINIMU_ERROR != CGEQ_ERROR
#error "fix definition of MINIMU_ERROR in minimu.h"
#endif
#if MINIMU_USER_END != CGEQ_USER_END
#error "fix definition of MINIMU_USER_END in minimu.h"
#endif
#if MINIMU_CONSTRAINT_ERROR != CGEQ_CONSTRAINT_ERROR
#error "fix definition of MINIMU_CONSTRAINT_ERROR in minimu.h"
#endif


#ifdef TEST

#define NSPINS 20
static int test_normalize(int argc, char *argv[])
{
	int i, repeat;
	FLOAt tmp, errmax;
	int bad= 0;
	FLOAt spins[3*NSPINS];
	static const FLOAt errlimit= 1e-6;

	for(repeat= 0; repeat<10; repeat++) {
		for(i= 0; i<3*NSPINS; i++) spins[i]= badrandom()-0.5;
		if(DECORATE(minimu_normalize)(3*NSPINS, spins, 0)) {
			bad++;
			puts("minimu_normalize() failed");
		}
		errmax= 0;
		for(i= 0; i<3*NSPINS; i+= 3) {
			tmp= FABS(
				spins[i]*spins[i] +
				spins[i+1]*spins[i+1] +
				spins[i+2]*spins[i+2] -
				1
			);
			if(errmax<tmp) errmax= tmp;
		}
		if(errmax > errlimit) {
			bad++;
			printf(
				"norm error %e exceeds limit %e\n",
				errmax, errlimit
			);
		}
	}

	return !!bad;
} /* test_normalize() */

static int test_project(int argc, char *argv[])
{
	int i, repeat;
	FLOAt tmp, errmax;
	FLOAt spins[3*NSPINS], delta[3*NSPINS];
	static const FLOAt errlimit= 1e-6;
	int bad= 0;

	for(repeat= 0; repeat<10; repeat++) {
		for(i= 0; i<3*NSPINS; i++) spins[i]= badrandom()-0.5;
		DECORATE(minimu_normalize)(3*NSPINS, spins, 0);
		for(i= 0; i<3*NSPINS; i++) delta[i]= badrandom()-0.5;
		DECORATE(minimu_project)(3*NSPINS, spins, delta, 0);
		errmax= 0;
		for(i= 0; i<3*NSPINS; i+= 3) {
			tmp= FABS(
				spins[i]*delta[i] +
				spins[i+1]*delta[i+1] +
				spins[i+2]*delta[i+2]
			);
			if(errmax<tmp) errmax= tmp;
		}
		if(errmax>errlimit) {
			bad++;
			printf(
				"minimu_project error %e exceeds limit %e\n",
				errmax, errlimit
			);
		}
	}

	return !!bad;
} /* test_project() */

#include "slab_problem.c"
/* Defines:
 *   struct slab_data,
 *   slab_field(), slab_init() slab_gnuplot_vectordump()
 */

static int slab_test(int argc, char *argv[])
{
	struct slab data= {
		.nx= 30, .ny= 14,
		.dx= 20e-7, .dy= 20e-7,	// cm
		.A= 1e-6,		// ergs/cm
		.Ku= 5.*800/2,		// ergs/cc (for Hk= 5 Oe)
		.Ms= 800,		// emu/cc
		.t= 40e-7		// cm
	};
	int ns= data.nx * data.ny;	// number of spins
	int n3= 3*ns;			// number of spin components
	FLOAt *s;			// bucket of numbers
	int bad= 0;
	int status;
	int breakout;
	FLOAt r2, r2goal;
	char  *magname= 0, *cglogname= 0;
	FILE    *cglog= NULL;
	FILE  *magfile= NULL;
	FILE   *gpfile= NULL;
	FILE *gpcgfile= NULL;
	int insist;
	static const int max_insist= 10;

	s= malloc((1+4)*n3*sizeof(*s));
	if(0==s) return -1;

	if(verbosity) {
		cglog= openlogfile("slab-cg", &cglogname);
		magfile= openlogfile("slab-mag", &magname);
		gpfile= openlogfile("slab-gp", 0);
		gpcgfile= openlogfile("slab-cg-gp", 0);
	}

	if(verbosity && gpfile) fprintf(gpfile,
		"set xrange [%e : %e]\n"
		"set yrange [%e : %e]\n",
		-1.0, 0.0+data.nx,  -1.0, 0.0+data.ny
	);
	if(verbosity && gpfile) fprintf(gpfile,
		"unset border\n"
		"unset xtics\n"
		"unset ytics\n"
		"unset key\n"
		"set size ratio %e\n",
		1.0*data.ny / data.nx
	);
	if(verbosity && gpcgfile) fputs(
		"# Run through gnuplot to trace the convergence\n"
		"# rate of micromagnetic conjugate gradients\n\n"
		"set logscale y\n"
		"set style data linespoints\n\n"
		,gpcgfile
	);

	if(0 != slab_init_matrix(&data)) {
		puts("slab_test: matrix initialization failed");
		bad++;
		goto out;
	}
	if(0 != slab_init_state(s, &data)) {
		puts("slab_test: state initialization failed");
		bad++;
		goto out;
	}

	r2goal= ns*1e-6;		// (1 mOe)^2
	for(insist= 0; insist<max_insist; insist++) {
		status= DECORATE(minimu)(
			ns, s, s+n3,
			r2goal, &r2, 50,
			slab_field, 0,
			&data,
			cgeq_simple_log_hook, cglog
		);
		breakout= (MINIMU_OK == status) || (MINIMU_ROUNDOFF == status);

		if(verbosity) {
			if(cglog) fprintf(cglog,
				"# cgeq= %#x, r2= %e\n\n\n", status, r2
			);
			if(magfile) slab_gnuplot_vectordump(magfile, s, &data);

			if(gpfile) fprintf(gpfile,
				"set title 'round %d %s, r2= %e\n",
				insist,
				(breakout ? "(final)" : ""),
				r2
			);
			if(gpfile) fprintf(gpfile,
				"\nplot '%s' index %d with vectors\n"
				"pause -1\n",
				magname, insist
			);
		} // if(verbosity)

		if(breakout) break;
	} // for(insist)

	if(verbosity && (max_insist == insist)) {
		if(gpfile) fprintf(gpfile, "set title 'final, r2= %e\n", r2);
		if(gpfile) fprintf(gpfile,
			"\nplot '%s' index %d with vectors\npause -1\n",
			magname, insist-1
		);
	}

	if(gpcgfile) fprintf(gpcgfile,
		"plot '%s'\n"
		"pause -1\n"
		, cglogname
	);

	/* it really should converge */
	if(!!status) bad++;
	if(r2 >= r2goal) bad++;

	out:
	free(s);
	slab_release_matrix(&data);
	if(magname)   free(magname);
	if(cglogname) free(cglogname);
	if(gpfile)  fclose(gpfile);
	if(magfile) fclose(magfile);
	if(cglog)   fclose(cglog);
	return !!bad;
}	/* slab_test() */

int (*tests[])(int argc, char *argv[])= {
	test_normalize, test_project,
	slab_test,
	0
};
#endif /* TEST */
