/* $Id: minimu.h,v 1.12.2.1 2007/04/08 23:32:01 pa Exp $
 * Energy minimization, unconstrained micromagnetics
 * (unconstrained besides the usual condition || s[i] || == 1, that is)
 *
 * Copyright (C) 2007  Seagate Technology
 * See terms at the end of this file.
 */
#ifndef MINIMU_H
#define MINIMU_H

#ifdef __cplusplus
extern "C" {
#endif

/* User must supply a "field" computation routine.
 * Read on below for why the word "field" is in quotes.
 */
typedef void (*minimu_field_proc)(
	int n,			/* number of spins */
	const double s[],	/* s[3*n], == [sx,sy,sz, sx,sy,sz ...] */
	double h[],		/* h[3*n], to store fields */
	void *context		/* pointer to anything they need */
);


/* User MAY supply a "preconditioning" routine.
 */
typedef void (*minimu_precond)(
	int n,			/* number of spins */
	const double s[],	/* s[3*n], the spins */
	const double h[],	/* h[3*n], "fields" */
	double delta_s[],	/* delta_s[3*n], output */
	void *context		/* pointer to anything they need */
);


/* Having written a field routine and maybe a preconditioning routine,
 * user can call the following to find a micromagnetic equilibrium state.
 * This state will be a local minimum of the energy.  It may be
 * metastable but won't be unstable.
 */
extern int minimu(
	int n,			/* number of spins */
	double s[],		/* s[3*n], guess, replaced by solution */
	double scratch[],	/* scratch[12*n] or NULL, work vector */
	double rtorq2,		/* exit criterion, squared residual torques */
	double *rtorq2_actual,	/* output, residual actually achieved */
	int max_iterations,	/* iteration limit */
	minimu_field_proc get_h,/* field computation routine */
	minimu_precond get_ds,	/* optional, pass NULL to do without */
	void *context,		/* forwarded to the two callbacks */
	/*
	* Next two arguments are normally NULL.  See the discussion
	* at the bottom of file "cgeq.h".
	*/
	int (*logging_hook)(int opcode, void *hook_context, ...),
	void *hook_context
);

/* Single-precision versions */
typedef void (*minimu_field_proc_f)(
	int n, const float s[], float h[], void *context
);
typedef void (*minimu_precond_f)(
	int n, const float s[], const float h[], float delta_s[], void *context
);
extern int minimu_f(
	int n, float s[], float scratch[],
	float rtorq2_goal, float *rtorq2, int max_iterations,
	minimu_field_proc_f get_h, minimu_precond_f get_ds,
	void *context,
	int (*logging_hook)(int opcode, void *hook_context, ...),
	void *hook_context
);

/* return codes from minimu() and minimu_f() */
#define MINIMU_OK	00	/* converged, no error */
#define MINIMU_ROUNDOFF	01	/* probably converged, can't improve */
#define MINIMU_ITMAX	02	/* iteration count exceeded, try again */
#define MINIMU_ERROR	03	/* something went wrong */
#define MINIMU_USER_END	04	/* terminated by logging hook */
#define MINIMU_CONSTRAINT_ERROR	05	/* bad initial state */



#ifdef __cplusplus
}
#endif

/* Why the quotes around the word "field" ?
 * minimu() finds local minima of the the energy E{s}.
 * To do this, it needs the partial derivatives of the energy,
 * in the form
 *     h[i] = - \partial E / \partial s[i] .
 *
 * These are proportional to the effective fields, but the
 * proportionality constants may have a few quirks.  For example in a
 * typical recording simulation each h[i] is the product Heff[i]*MsV[i]
 * of the effective field *and moment (Ms*V) * of grain i.  Leaving out
 * the moment factors won't work (unless all grains are identical).
 *
 * The mathematical condition on h[] is that its cross-derivatives be
 * equal, to machine precision:
 *
 *     \partial h[i] / \partial s[j] == \partial h[j] / \partial s[i] .
 *
 * This is the nonlinear version of reciprocity.  It is equivalent
 * to the statement that the effective fields are derivable from an energy.
 * The computed "fields" MUST obey this identity or the algorithm
 * won't converge.
 *
 * ----
 * The preconditioning routine must be a linear mapping from fields
 * to spin displacement,
 *
 *     delta_s[] = M * h[]
 *
 * where M is a 3n x 3n symmetric positive-definite matrix.
 * M can be a function of s[], but not of h[].  Preconditioning can
 * improve the convergence rate.
 *
 * In a recording simulation, a simple preconditioning operation would
 * be to divide out the moments from h[], i.e. set delta_s of grain i
 * to h[] / MsV[i].  The matrix M corresponding to this is diagonal,
 * with entries equal to the reciprocal volumes.  It is symmetric
 * because it is diagonal, and it is positive-definite because the
 * volumes are positive.
 */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#endif /* MINIMU_H */
