/* Sample use of micromagnetic routines from C.
 * Compute a classical Landau-Lifshitz wall profile.
 * The system is a linear spin chain with exchange,
 * bi-axial anisotropy, no external field and no demag.
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

/* Structure that holds what we need for the field computation */
struct llw_params {
	double A;		/* exchange stiffness */
	double Ku;		/* anisotropy */
	double Ko;		/* "orthorhombic" anisotropy */
	double ea[3];		/* unit vector along easy axis */
	double ma[3];		/* unit vector along medium axis */
	double left_spin[3];	/* unit vector, boundary condition */
	double right_spin[3];	/* unit vector, boundary condition */
	double *save_energy;	/* place to store E, or NULL */
};

/* field computation routine */
void get_fields(int n, const double s[], double h[], void *context)
{
	const struct llw_params *p= context;	/* get our parameters */
	const double *ea= p->ea;
	const double *ma= p->ma;

	int i, k;
	const int n3= n*3;
	double E;
	/* convenient views of the spin and field arrays
	 * s3[i][k], h3[i][k] == component k at site i
	 * (see C99 rationale 6.7.5.3, p. 87, lines 12--16)
	 */
	const double (* const s3)[3]= ((const double (*)[3]) s);
	double (* const h3)[3]= ((double (*)[3]) h);

	double tmp, tmp1;

	/* zero the fields */
	E= 0;
	for(i= 0; i<n3; i++) h[i]= 0;
	/* now go through all the energy terms and add their
	 * negative gradients to h3[][]
	 */

	/* exchange, interior */
	for(i= 0; i<n-1; i++) for(k= 0; k<3; k++) {
		tmp= (s3[i+1][k] - s3[i][k]);
		tmp1= tmp * (p->A);
		h3[i][k]+= tmp1; h3[i+1][k]-= tmp1;
		E+= 0.5*tmp*tmp1;
	}
	/* exchange, endpoints */
	for(k= 0; k<3; k++) {
		tmp= s3[0][k] - p->left_spin[k];
		tmp1= tmp * (p->A);
		h3[0][k]-= tmp1; E+= 0.5*tmp*tmp1;
		tmp= p->right_spin[k] - s3[n-1][k];
		tmp1= tmp * (p->A);
		h3[n-1][k]+= tmp1; E+= 0.5*tmp*tmp1;
	}

	/* anisotropy */
	for(i= 0; i<n; i++) {
		/* easy axis */
		for(tmp= 0, k= 0; k<3; k++) tmp+= s3[i][k]*ea[k];
		tmp1= tmp * (p->Ku);
		for(k= 0; k<3; k++) h3[i][k]+= tmp1*ea[k];
		E+= 0.5*(p->Ku -tmp*tmp1);
		/* medium axis */
		for(tmp= 0, k= 0; k<3; k++) tmp+= s3[i][k]*ma[k];
		tmp1= tmp * (p->Ko);
		for(k= 0; k<3; k++) h3[i][k]+= tmp1*ma[k];
		E+= 0.5*(p->Ko - tmp*tmp1);
	}
	/* and we're done! */
	if(p->save_energy) *p->save_energy= E;
} /* get_fields() */


/* stolen from the rand(3) manpage, to avoid platform dependencies */
static unsigned long next = 1;
double badrandom(void) {
	next = next * 1103515245 + 12345;
	return ((unsigned)(next/65536) % 32768)/32768.0;
}

/* FIXME show how to write a preconditioning routine.
 * We won't use one for now, just pass NULLs to the micromag routines.
 */

#include "minimu.h"
#include "minimu_push.h"
#include "cgeq.h"
/* cgeq.h needed only for the cgeq_simple_log_hook() */

#include <stdio.h>
#include <math.h>
#include <string.h>

int main(int argc, char *argv[])
{
	/* energy */
	double E;

	/* material parameters */
	const struct llw_params p= {
		144.0,		/* .A */
		1.0,		/* .Ku */
		0.1,		/* .Ko */
		{0, 0, 1},	/* .ea[] */
		{1, 0, 0},	/* .ma[] */
		{0, 0,-1},	/* .left_spin[] */
		{0, 0, 1},	/* .right_spin[] */
		&E,		/* .save_energy */
	};

	/* system size */
	const int nspins= 100;

	/* spin arrays */
	double  s[nspins][3];	/* main array */
	double s0[nspins][3]; 	/* constraint reference */
	double  w[nspins];	/* weights */

	/* miscellaneous */
	double resid2, ell, ell2, rlambda, ell2_goal;
	int status, nrepeats;
	int i, k;

	/* output and log files */
	FILE *spinfile= fopen("llw-dat.out", "w");	/* skip error checks */
	FILE *gnuplotfile= fopen("llw-gp.out", "w");
	FILE *cglogfile= fopen("llw-cg.out", "w");
	setlinebuf(cglogfile);			/* so we can "tail -f" it */

	/* prepare some initial state */
	{
		double tmp, tmp0, tmp1, tmp2;

		for(i= 0; i<nspins; i++) {
			tmp0= ((double) 4*i*(nspins-i-1))/(nspins*nspins);
			tmp2= ((double) 2*i-nspins)/nspins;
			tmp1= 0.2*tmp0;
			tmp= 1.0/sqrt(tmp0*tmp0 + tmp1*tmp1 + tmp2*tmp2);
			s[i][0]= tmp*tmp0;
			s[i][1]= tmp*tmp1;
			s[i][2]= tmp*tmp2;
		}
	}

	/* Ordinary micromagnetic minimization
	 *
	 * We make up to five calls to minimu(), each with
	 * a 200 iteration limit, and we break out as soon as
	 * the error goal is met or roundoff is reached.
	 */
	puts("Unconstrained micromagnetic minimum...\n");
	fputs("# Unconstrained micromagnetic minimization\n#\n", cglogfile);
	for(nrepeats= 1; nrepeats<=5; nrepeats++) {
		status= minimu(
			nspins, &s[0][0], /* number of spins, spin array */
			0,		  /* no scratch space */
			1.0e-10, &resid2, /* errgoal, errreturn */
			200,		  /* maxiter */
			get_fields,	  /* field routine */
			0, 		  /* no preconditioning routine */
			/* p is cast to discard the const qualifier */
			(struct llw_params *) &p,
			/* simple conjugate gradient logging,
			 * see bottom of cgeq.h
			 */
			cgeq_simple_log_hook, cglogfile
		);
		fprintf(cglogfile, "# minimu() returns %d\n\n", status);
		if((MINIMU_OK==status) || (MINIMU_ROUNDOFF==status)) break;
	}
	printf("minimu() returns %d after pass %d\n", status, nrepeats);
	printf("final residual: %f\n", sqrt(resid2));
	printf("final energy: %f\n", E);

	/* save results for plotting */
	fprintf(spinfile,
		"# Unconstrained micromagnetic minimum\n#\n"
		"# %2s  %10s  %10s  %10s\n",
		"n", "mx", "my", "mz"
	);
	for(i= 0; i<nspins; i++) fprintf(spinfile,
		"% 4d  % 10.5f  % 10.5f  % 10.5f\n",
		i, s[i][0], s[i][1], s[i][2]
	);
	fputs(
		"# [Linux] pass this file to gnuplot "
		"to plot the Landau-Lifshitz\n"
		"# wall profile generated by example.c\n\n"
		"set data style linespoints\n"
		"set key left\n"
		"set title 'Landau-Lifshitz wall profile'\n"
		"plot 'llw-dat.out' index 0 using 1:4, \\\n"
		"    'llw-dat.out' index 0 using 1:3, \\\n"
		"    'llw-dat.out' index 0 using 1:2\n"
		"pause -1\n"
		, gnuplotfile
	);

	/* Setup for constrained minimization.
	 * Save the minimum obtained above to use as a reference state.
	 */
	memcpy(s0[0], s[0], 3*nspins*sizeof(s0[0][0]));

	/* initialize the weights */
	for(i= 0; i<nspins; i++) w[i]= 1;

	/* Perturb the solution.  No need to keep it normalized, the
	 * call to minimu_push_make_feasible() below will take care of
	 * that.
	 */
	puts("\nPerturbing the micromagnetic minimum...");
	for(i= 0; i<nspins; i++) for(k= 0; k<3; k++) {
		s[i][k]+= 0.01*(badrandom()-0.5);
	}

	/* Pick a squared Euclidean distance, here 1e-4, and
	 * project the perturbed solution to the feasible manifold.
	 * This routine may fail, but if so it will tell us what
	 * squared Euclidean length it managed to reach.
	 */
	puts("Projecting to the feasible manifold...");
	status= minimu_push_make_feasible(
		nspins, s[0],	/* number of spins, spin arrays */
		s0[0], w,	/* reference spins, weights */
		1e-4,		/* the L^2 that we want */
		&ell2		/* return value for the L^2 we get */
	);
	printf("minimu_push_make_feasible() returns %d\n", status);
	printf("Euclidean distance achieved: %f\n", ell= sqrt(ell2));

	/* Minimize subject to the Euclidean constraint.
	 * Note that for the squared distance we pass the value
	 * returned by minimu_push_make_feasible() above.
	 */
	puts("\nConstrained minimization...");
	fputs("\n\n# Constrained minimization\n#\n", cglogfile);
	status= minimu_push(
		nspins, &s[0][0],	/* number of spins, spin array */
		0,			/* no scratch space */
		1e-10, &resid2,		/* errgoal, errreturn */
		1000,			/* errgoal */
		get_fields,		/* field routine */
		0,			/* no preconditioning routine */
		&s0[0][0], w,		/* reference spins, weights */
		ell2,			/* squared distance */
		&rlambda,		/* output, Lagrange multiplier */
		/* parameters, discard the const */
		(struct llw_params *) &p,
		/* simple logging */
		cgeq_simple_log_hook, cglogfile
	);
	printf("minimu_push() returns %d\n", status);
	fprintf(cglogfile, "# minimu_push() returns %d\n\n", status);
	printf("final residual: %f\n", sqrt(resid2));
	printf("final energy: %f\n", E);
	printf("Lagrange multiplier: %f\n", rlambda);

	/* output, displacement from reference */
	fprintf(spinfile,
		"\n\n"
		"# Small displacement due to constraint, l^2= %f\n#\n"
		"# %2s  %10s  %10s  %10s\n",
		ell2,
		"n", "dmx", "dmy", "dmz"
	);
	for(i= 0; i<nspins; i++) fprintf(spinfile,
		"% 4d  % 10.5f  % 10.5f  % 10.5f\n",
		i,
		(s[i][0]-s0[i][0])/ell,
		(s[i][1]-s0[i][1])/ell,
		(s[i][2]-s0[i][2])/ell
	);
	fputs(
		"\n"
		"set title 'Lowest eigenvector'\n"
		"plot 'llw-dat.out' index 1 using 1:4, \\\n"
		"    'llw-dat.out' index 1 using 1:3, \\\n"
		"    'llw-dat.out' index 1 using 1:2\n"
		"pause -1\n"
		, gnuplotfile
	);

	/* Second constrained minimization.
	 * Attempt to increase the distance.
	 */
	ell= 0.5;
	printf("\nAttempting to increase distance to %f ....\n", ell);
	status= minimu_push_make_feasible(
		nspins, s[0], s0[0], w, ell*ell, &ell2
	);
	printf("minimu_push_make_feasible() returns %d\n", status);
	printf("Distance achieved: %f\n", ell= sqrt(ell2));

	puts("\nSecond constrained minimization.");
	fputs("\n\n# Second constrained minimization\n#\n", cglogfile);
	status= minimu_push(
		nspins, &s[0][0], 0, 1e-10, &resid2, 1000,
		get_fields, 0, &s0[0][0], w, ell2, &rlambda,
		(struct llw_params *) &p,
		cgeq_simple_log_hook, cglogfile
	);
	printf("minimu_push() returns %d\n", status);
	fprintf(cglogfile, "# minimu_push() returns %d\n\n", status);
	printf("final residual: %f\n", sqrt(resid2));
	printf("final energy: %f\n", E);
	printf("Lagrange multiplier: %f\n", rlambda);

	fprintf(spinfile,
		"\n\n# Constrained micromagnetic minimum, l^2= %f\n#\n"
		"# %2s  %10s  %10s  %10s\n",
		ell2,
		"n", "mx", "my", "mz"
	);
	for(i= 0; i<nspins; i++) fprintf(spinfile,
		"% 4d  % 10.5f  % 10.5f  % 10.5f\n",
		i, s[i][0], s[i][1], s[i][2]
	);
	fputs(
		"\n\n"
		"set title 'Wall distorted by constraint'\n"
		"plot 'llw-dat.out' index 2 using 1:4, \\\n"
		"    'llw-dat.out' index 2 using 1:3, \\\n"
		"    'llw-dat.out' index 2 using 1:2\n"
		"pause -1\n"
		, gnuplotfile
	);

	fclose(cglogfile);
	fclose(gnuplotfile);
	fclose(spinfile);

	/* How to safely scan L^2.
	 * We start over close to the unconstrained equilibrium and
	 * solve constrained problems at increasing distance.
	 * We try to increase L^2 to 3.0 in steps of 0.5 .  These are
	 * big steps so minimu_push_make_feasible() will probably fail
	 * and we will have to recover.
	 *
	 * This time just report the energy and Lagrange multiplier,
	 * don't save the data.
	 *
	 * Note: we are not going to find saddle points in this system
	 * because of the boundary conditions we used.  The constrained
	 * energy should just increase monotonically, until we push
	 * L^2 too close to 4.  At that point probably an end spin will
	 * tear off the boundary condition, or maybe a reversal will form
	 * and give a state with three walls.  Either way, the energy
	 * will drop discontinuously with L^2.
	 * I'm not particularly interested in the details.
	 */
	/* Return near the reference point */
	for(i= 0; i<nspins; i++) for(k= 0; k<3; k++) {
		s[i][k]= s0[i][k] + 0.01*(badrandom() - 0.5);
	}

	printf(
		"\n\n%7s  %9s  %9s  %10s  %s\n",
		"L^2", "lambda", "E", "resid", "status"
	);
	/* Special code to redo the case L^2= 0.
	 * Call minimu() with a high residual limit and an iteration limit
	 * of zero, just to recompute the energy and residuals without
	 * doing anything else.
	 */
	ell2= 0;
	status= minimu(
		nspins, &s0[0][0], 0, 1e99, &resid2, 0,
		get_fields, 0, (struct llw_params *)&p, 0, 0
	);
	printf("% 6.4f  % 9.5f  % 9.5f  % 10.3e  %d\n",
		ell2, rlambda, E, resid2, status
	);

	/* The ramp on L^2, 0 < L^2 < 3, taking into account the
	 * possible failure of minimu_push_make_feasible().
	 */
	for(i= 1; i<= 6; i++) {
		ell2_goal= 0.5*i;
		while(fabs(ell2-ell2_goal)>1e-6) {
			/* may need multiple attempts */
			status= minimu_push_make_feasible(
				nspins, &s[0][0], &s0[0][0], w,
				ell2_goal, &ell2
			);
			for(k= 0; k<5; k++) {
				status= minimu_push(
					nspins, &s[0][0], 0,
					1e-10, &resid2, 200,
					get_fields, 0,
					&s0[0][0], w, ell2, &rlambda,
					(struct llw_params *)&p,
					0, 0	/* no CG logging this time */
				);
				if(
					MINIMU_OK==status ||
					MINIMU_ROUNDOFF==status
				) break;
			}
			printf("% 6.4f  % 9.5f  % 9.5f  % 10.3e  %d\n",
				ell2, rlambda, E, resid2, status
			);
		}
	}

	return 0;
}; /* main() */
