/* $Id: minimu_push.c,v 1.37.2.1 2007/04/08 23:32:01 pa Exp $
 * Energy minimization,  micromagnetic constraints plus additional
 * constraint to push the solution away from a given state (usually a
 * micromagnetic minimum).
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#include <stdlib.h>
#include "minimu_push.h"

#ifdef TEST
#include "test-rig.c"

/* tests need minimu.c and cgeq.c without their own test harnesses */
#undef TEST
#include "cgeq.c"
#include "cgeq_simplehook.c"
#include "minimu.c"
#define TEST
#endif /* TEST */

#include "config.h"

#include <math.h>
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif /* HAVE_FLOAT_H */
#include <string.h>
#include "cgeq.h"

/* STOP
 * Read the following header first.
 */
#include "minimu_P.h"

/* cgeq callbacks.  Same as in minimu.c .
 */
struct minimu_push_wrap {
	struct DECORATE(minimu_wrap) mmu;
	struct constraint {
		FLOAt l2;	 /* target for < (2\tan(\theta/2))^2 > */
		FLOAt lambda;	 /* output, the Lagrange multiplier */
		const FLOAt *s0; /* s0[3*n], reference spins */
		const FLOAt *w;	 /* w[n], weights */
		FLOAt *scratch;	 /* scratch[n] */
	} mmu_push;
};
/* This struct is passed, as a (void *), to routines in minimu_P.h
 * that expect a struct minimu_wrap.  This is ok because the first
 * component is the minimu_wrap they expect.  By the C standard,
 * pointers to a struct also point to the first component.
 * (In C99: 6.7.2.1, Semantics, paragraph 13.)
 */

/* default preconditioner, uses the weights in the struct constraint */
static void precondition(
	    int n3, const FLOAt s[], const FLOAt h[], FLOAt p[], void *context
)
{
	struct constraint *c= &(
		(struct minimu_push_wrap *) context
	)->mmu_push;
	const FLOAt *w= c->w;
	int i, n= n3/3;
	const FLOAt *rh;
	FLOAt *rp, tmp;
	for(
		i= 0, rh= h, rp= p;
		i<n;
		i++, rh+= 3, rp+= 3
	) {
		tmp= 1/w[i];
		rp[0]= rh[0]*tmp;
		rp[1]= rh[1]*tmp;
		rp[2]= rh[2]*tmp;
	}
}

/* Step size picker.  Take l2 into account. */
static int stepsize(
	int n, const FLOAt s[], const FLOAt ds[],
	FLOAt *s0, FLOAt *smax,
	void *context
)
{
	struct constraint *c= &(
		(struct minimu_push_wrap *)context
	)->mmu_push;
	FLOAt tmp;

	if(DECORATE(minimu_stepsize)(n, s, ds, s0, smax, 0)) {
		return -1;
	}
	tmp= c->l2;
	if(tmp<1) {
		tmp= SQRT(tmp);
		*s0*= tmp;
		*smax*= tmp;
	}
	return 0;
}

static int normalize(int n3, FLOAt s[], void *context)
{
	int i;
	int n= n3/3;
	int status= 0;
	struct constraint *c= &(
		(struct minimu_push_wrap *)context
	)->mmu_push;
	FLOAt lambda, gamm0,gamm1, sigm0,sigm1, mu;
	FLOAt mumin, mumax;
	FLOAt tmp;
	const FLOAt *w= c->w;
	FLOAt W;
	FLOAt *d2= c->scratch;
	FLOAt *rs; const FLOAt *rs0;	/* "rovers" */
	FLOAt dsx,dsy,dsz;

	/* bring them to unit length */
	if(DECORATE(minimu_normalize)(n3, s, 0)<0) return -1;

	/* 1st pass, calculate < d^2 > and < d^2*(4-d^2) >
	 * Use gamm0 and gamm1 as temporary accumulators.
	 *
	 * Also keep track of Max(d^2) and Max(4-d^2) == 4-Min(d^2)).
	 */
	gamm0= gamm1= W= 0;
	mumin= 4;			/* for Min(d^2) */
	mumax= 0;			/* for Max(d^2) */
	for(i= 0, rs= s, rs0= c->s0; i<n; i++, rs+= 3, rs0+= 3) {
		dsx= rs[0]-rs0[0];
		dsy= rs[1]-rs0[1];
		dsz= rs[2]-rs0[2];
		tmp= dsx*dsx + dsy*dsy + dsz*dsz;
		d2[i]= tmp;
		if(tmp<mumin) mumin= tmp;	/* mumin == Min(d^2) */
		if(tmp>mumax) mumax= tmp;	/* mumax == Max(d^2) */
		W+= w[i];
		gamm0+= w[i]*tmp;;
		gamm1+= w[i]*tmp*(4-tmp);
	}
	mumin= 4-mumin;

	/* At this point,
	 *   mumin = Max(4-d^2)
	 *   mumax = Max(d^2)
	 *
	 * The bounds for mu are -1/Max(4-d^2) <= mu <= 1/Max(d^2)
	 * so I need to invert mumin and mumax.  We'd better not get a
	 * division by zero, because it would mean either that
	 * all spins are at zero degrees (mumax == 0) or 180 degrees
	 * (mumin==0), in which case we're utterly stuck because the
	 * user passed us s[] == s0[] or -s0[].
	 */
	if(mumin<=0  ||  mumax<=0) return -1;
	mumin= -1./mumin;
	mumax= 1./mumax;

	/* Set mu == coefficient by which we want to adjust distances,
	 * d^2 <--- d^2 + mu * d^2*(4-d^2)
	 */
	if(0==W) return -1;
	if(0==gamm1) return -1;
	mu= (W*c->l2 - gamm0)/gamm1;
	if(mu < mumin) {
		mu= mumin;
		status= -1;
	}
	if(mu > mumax) {
		mu= mumax;
		status= -1;
	}

	/* 2nd pass, shift the spins */
	for(i= 0, rs= s, rs0= c->s0; i<n; i++, rs+= 3, rs0+= 3) {
		tmp= d2[i];		/* d_0^2 */
		gamm0= 1 - tmp/2;	/* \cos(\theta_0) */
		sigm0= tmp*(4-tmp);	/* 4 \sin^2(\theta_0) */

		tmp+= mu*sigm0;		/* d_1^2 */
		gamm1= 1 - tmp/2;	/* \cos(\theta_1) */
		sigm1= tmp*(4-tmp);	/* 4 \sin^2(\theta_1) */

		/* skip zero-divide from roundoff near angles 0 and pi,
		 * those spins wouldn't rotate anyway
		 */
		if(0>=sigm1) continue;

		lambda= gamm0 - gamm1*SQRT(sigm0/sigm1);
		rs[0]-= lambda*rs0[0];
		rs[1]-= lambda*rs0[1];
		rs[2]-= lambda*rs0[2];
	}

	/* final, bring back to unit length */
	return status + DECORATE(minimu_normalize)(n3, s, 0);
} /* normalize() */

/* Since the normalization is nontrivial, we export a function to do
 * it.  This is NOT suitable as a callback for cgeq().
 */
int DECORATE(minimu_push_make_feasible)(
	int n,
	FLOAt s[],
	const FLOAt s0[], const FLOAt w[],
	FLOAt l2, FLOAt *l2_actual
)
{
	int status= 0;
	struct minimu_push_wrap mmpw= {
		{
			0,		/* .get_field */
			0,		/* .get_steepest */
			0		/* .context */
		}, 	/* .mmu */
		{
			l2,		/* .l2 */
			0,		/* .lambda */
			s0,		/* .s0 */
			w,		/* .w */
			0		/* .scratch */
		}	/* .mmu_push */
	};
	mmpw.mmu_push.scratch= malloc(n*sizeof(*mmpw.mmu_push.scratch));
	if(0 == mmpw.mmu_push.scratch) return -1;

	status= normalize(n*3, s, &mmpw);

	/* Store the actual distance if they ask for it.
	 * If normalize() claims it worked we take its word for it;
	 * otherwise, we compute the distance directly.
	 */
	if(l2_actual) {
		if(0==status) {
			*l2_actual= l2;
		} else {
			int i;
			const FLOAt *rs0;
			FLOAt *rs, dsx,dsy,dsz, W, sum;
			W= sum= 0;
			for(
				i= 0, rs= s, rs0= s0;
				i<n;
				i++, rs+= 3, rs0+= 3
			) {
				dsx= rs[0]-rs0[0];
				dsy= rs[1]-rs0[1];
				dsz= rs[2]-rs0[2];
				W+= w[i];
				sum+= w[i]*(dsx*dsx + dsy*dsy + dsz*dsz);
			}
			if(W <= 0) *l2_actual= -1;
			else *l2_actual= sum/W;
		}
	}

	free(mmpw.mmu_push.scratch);
	return status;
} /* minimu_push_make_feasible() */


#ifdef TEST

#define NSPINS 30
static int test_normalize(int argc, char *argv[])
{
	int i, repeat;
	int status;
	FLOAt s[3*NSPINS], s0[3*NSPINS], w[NSPINS], scratch[NSPINS];
	struct minimu_push_wrap mmpw= {
		.mmu_push= {
			.s0= s0, .w= w, .scratch= scratch,
		}
	};
	int bad= 0;
	FLOAt l2, W;
	FLOAt err, tmp;
	FLOAt errlimit= 200*DECORATE(epsilon);

	for(repeat= 0; repeat<100; repeat++) {
		/* reference spins */
		for(i= 0; i<3*NSPINS; i++) {
			s0[i]= 2*badrandom()-1;
		}
		DECORATE(minimu_normalize)(3*NSPINS, s0, 0);

		/* test spins */
		for(i= 0; i<3*NSPINS; i++) {
			s[i]= s0[i] + (badrandom()-0.5);
		}

		/* weights, target */
		for(i= 0; i<NSPINS; i++) w[i]= badrandom();
		for(W= tmp= 0, i= 0; i<NSPINS; i++) {
			W+= w[i];
			tmp+= w[i]*(
				(s[3*i+0]-s0[3*i+0]) * (s[3*i+0]-s0[3*i+0]) +
				(s[3*i+1]-s0[3*i+1]) * (s[3*i+1]-s0[3*i+1]) +
				(s[3*i+2]-s0[3*i+2]) * (s[3*i+2]-s0[3*i+2])
			);
		}
		mmpw.mmu_push.l2= (tmp/W)*(1 + 0.1*badrandom());

		status= normalize(3*NSPINS, s, &mmpw);

		for(err= 0, i= 0; i<NSPINS; i++) {
			tmp= FABS(
				s[3*i+0]*s[3*i+0] +
				s[3*i+1]*s[3*i+1] +
				s[3*i+2]*s[3*i+2] -
				1
			);
			if(err<tmp) err= tmp;
		}
		if(status<0) {
			if(verbosity) {
				printf(
					"normalized, pass %d, soft failure\n",
					repeat
				);
			}
		} else {
			W= l2= 0;
			for(i= 0; i<NSPINS; i++) {
				tmp= (
					(s[3*i+0]-s0[3*i+0]) *
					(s[3*i+0]-s0[3*i+0])
				) + (
					(s[3*i+1]-s0[3*i+1]) *
					(s[3*i+1]-s0[3*i+1])
				) + (
					(s[3*i+2]-s0[3*i+2]) *
					(s[3*i+2]-s0[3*i+2])
				);
				W+= w[i];
				l2+= w[i]*tmp;
			}
			l2*= (1/W);
			tmp= FABS(l2 - mmpw.mmu_push.l2);
			if(err<tmp) err= tmp;
			if(err>errlimit) {
				if(verbosity && 0==bad) puts("");
				bad++;
				printf(
					"normalize pass %d, "
					"error %e > limit %e\n",
					repeat, err, errlimit
				);
			}
		}
	} /* for(repeat) */

	return bad;
} /* test_normalize() */

# endif /* TEST */


/* tangent projection */
static void project(int n3, const FLOAt s[], FLOAt p[], void *context)
{
	int n= n3/3;
	struct constraint *c= &(
		(struct minimu_push_wrap *) context
	)->mmu_push;
	const FLOAt *rs, *rs0;
	FLOAt *rp;
	const FLOAt *w= c->w;
	FLOAt dsx,dsy,dsz, tmp, W, ps0, denom, lambda;
	int i;

	DECORATE(minimu_project)(n3, s, p, 0);

	if(c->l2 <= 0) return;	/* but they really shouldn't pass l2==0 */

	W= ps0= denom= 0;
	for(
		i= 0, rs= s, rs0= c->s0, rp= p;
		i<n;
		i++, rs+= 3, rs0+= 3, rp+= 3
	) {
		W+= w[i];
		dsx= rs[0]-rs0[0]; dsy= rs[1]-rs0[1]; dsz= rs[2]-rs0[2];
		ps0-= w[i]*(dsx*rp[0] + dsy*rp[1] + dsz*rp[2]);
		tmp= dsx*dsx + dsy*dsy + dsz*dsz;
		denom+= w[i] * tmp*(4-tmp);
	}
	if(W <= 0) return;		/* an error, but it's their fault */

	if(denom <= 0) return;	/* ditto */
	lambda= 4*ps0/denom;

	for(
		i= 0, rs= s, rs0= c->s0, rp= p;
		i<n;
		i++, rs+= 3, rs0+= 3, rp+= 3
	) {
		rp[0]+= lambda*(rs[0]-rs0[0]);
		rp[1]+= lambda*(rs[1]-rs0[1]);
		rp[2]+= lambda*(rs[2]-rs0[2]);
	}

	DECORATE(minimu_project)(n3, s, p, 0);
} /* project() */

#ifdef TEST

static int test_project(int argc, char *argv[])
{
	FLOAt s0[3*NSPINS], sa[3*NSPINS], sb[3*NSPINS];
	FLOAt  p[3*NSPINS];
	FLOAt w[NSPINS], scratch[NSPINS];
	struct minimu_push_wrap mmpw= {
		.mmu_push= {
			.s0= s0, .w= w, .scratch= scratch,
		}
	};
	int i, repeat;
	FLOAt tmp, err;
	const FLOAt errlimit= DECORATE(epsilon)*10;
	const FLOAt h= SQRT(DECORATE(epsilon));
	int status;
	int bad= 0;

	for(repeat= 0; repeat<100; repeat++) {
		/* target distance squared */
		mmpw.mmu_push.l2= 2*badrandom();

		/* weights */
		for(i= 0; i<NSPINS; i++) w[i]= badrandom();

		/* reference spins */
		for(i= 0; i<3*NSPINS; i++) {
			s0[i]= badrandom()-0.5;
		}
		DECORATE(minimu_normalize)(3*NSPINS, s0, 0);

		/* Base spins.
		* Don't expect to hit the requested distance, just
		* accept what make_feasible() manages.
		*/
		for(i= 0; i<3*NSPINS; i++) {
			sa[i]= s0[i] + (badrandom()-0.5);
		}
		status= DECORATE(minimu_push_make_feasible)(
			NSPINS, sa, s0, w,
			mmpw.mmu_push.l2, &mmpw.mmu_push.l2
		);

		/* displacement and perturbed spins */
		for(i= 0; i<3*NSPINS; i++) {
			p[i]= badrandom()-0.5;
			sb[i]= sa[i] + p[i]*h;
		}

		/* normalize perturbed spins, tangent-project displacement,
		* and verify that they still match to O(h)
		*/
		status= normalize(3*NSPINS, sb, &mmpw);
		project(3*NSPINS, sa, p, &mmpw);
		for(i= 0, err= 0; i<3*NSPINS; i++) {
			tmp= FABS(sb[i] - sa[i] - p[i]*h);
			if(err<tmp) err= tmp;
		}
		if(err>errlimit) {
			bad++;
			printf(
				"project pass %d, error %e > limit %e\n",
				repeat, err, errlimit
			);
		}
	}
	return bad;
} /* test_project() */

#endif /* TEST */

/* coprojection */
static void coproject(
	int n3, const FLOAt s[], FLOAt h[], void *context
)
{
	int n= n3/3;
	struct constraint *c= &(
		(struct minimu_push_wrap *) context
	)->mmu_push;
	int i;
	const FLOAt *s0= c->s0, *w= c->w;
	FLOAt tmp, dot, W, sum, mu;
	FLOAt dsx,dsy,dsz;
	const FLOAt *rs, *rs0;  FLOAt *rh;

	/* preprojection */
	DECORATE(minimu_project)(n3, s, h, 0);
	if(c->l2 <= 0) return;	/* but they shouldn't pass l2==0 */

	/* dot product (h,s0) */
	for(dot= 0, i= 0; i<n3; i++) {
		dot+= h[i]*s0[i];
	}

	/* weight normalization */
	for(W= 0, i= 0; i<n; i++) {
		W+= w[i];
	}
	if(W<=0) return;		/* an error, but it's their fault */
	sum= 0;
	for(i= 0, rs= s, rs0= s0;   i<n;   i++, rs+= 3, rs0+= 3) {
		dsx= rs[0]-rs0[0]; dsy= rs[1]-rs0[1]; dsz= rs[2]-rs0[2];
		tmp= dsx*dsx + dsy*dsy + dsz*dsz;
		sum+= w[i] * tmp*(4-tmp);
	}
	mu= 4*dot/sum;

	/* Lagrange multiplier */
	c->lambda= mu*W/2;

	/* constraint removal */
	for(
		i= 0, rs= s, rs0= s0, rh= h;
		i<n;
		i++, rs+= 3, rs0+= 3, rh+= 3
	) {
		rh[0]+= mu * w[i]*(rs[0]-rs0[0]);
		rh[1]+= mu * w[i]*(rs[1]-rs0[1]);
		rh[2]+= mu * w[i]*(rs[2]-rs0[2]);
	}

	/* postprojection */
	DECORATE(minimu_project)(n3, s, h, 0);
} /* coproject() */



#ifdef TEST

static int test_coproject(int argc, char *argv[])
{
	int bad= 0;
	int repeat, i;
	FLOAt s0[3*NSPINS], s[3*NSPINS];
	FLOAt w[NSPINS], scratch[NSPINS];
	struct minimu_push_wrap mmpw= {
		.mmu_push= {
			.s0= s0, .w= w, .scratch= scratch,
		}
	};
	FLOAt p[3*NSPINS], Tp[3*NSPINS];
	FLOAt h[3*NSPINS], Tdh[3*NSPINS];
	FLOAt err, errlimit= 10*DECORATE(epsilon);

	for(repeat= 0; repeat<100; repeat++) {
		/* reference spins */
		for(i= 0; i<3*NSPINS; i++) s0[i]= badrandom()-0.5;
		DECORATE(minimu_normalize)(3*NSPINS, s0, 0);

		/* weights, constraint */
		for(i= 0; i< NSPINS; i++) w[i]= badrandom();
		mmpw.mmu_push.l2= 3*badrandom();

		/* base spins */
		for(i= 0; i< 3*NSPINS; i++) {
			s[i]= s0[i] + (badrandom()-0.5);
		}
		normalize(3*NSPINS, s, &mmpw);

		/* displacement, field */
		for(i= 0; i< 3*NSPINS; i++) {
			p[i]= badrandom()-0.5;
			h[i]= badrandom()-0.5;
		}

		/* projections */
		memcpy( Tp, p, 3*NSPINS*sizeof( Tp[0]));
		memcpy(Tdh, h, 3*NSPINS*sizeof(Tdh[0]));
		project(3*NSPINS, s, Tp, &mmpw);
		coproject(3*NSPINS, s, Tdh, &mmpw);

		/* error */
		for(err= 0, i= 0; i<3*NSPINS; i++) {
			err+= (h[i]*Tp[i] - Tdh[i]*p[i]);
		}
		err= FABS(err);
		if(err>errlimit) {
			bad++;
			printf(
				"coproject pass %d, error %e > limit %e\n",
				repeat, err, errlimit
			);
		}
	} /* for(repeat) */

	return bad;
} /* test_coproject() */

# endif /* TEST */


int DECORATE(minimu_push)(
	int n,
	FLOAt s[], FLOAt scratch[],
	FLOAt rtorq2_goal, FLOAt *rtorq2, int max_iterations,
	DECORATE(minimu_field_proc) get_h,
	DECORATE(minimu_precond)    get_delta,
	const FLOAt s0[], const FLOAt w[], FLOAt l2, FLOAt *lambda,
	void *context,
	int (*logging_hook)(int opcode, void *logging_context, ...),
	void *logging_context
)
{
	int status;
	int get_own_scratch= (0==scratch);
	struct minimu_push_wrap mmpw= {
		{
			get_h,		/* .get_field */
			get_delta,	/* .get_steepest */
			context		/* .context */
		},	/* .mmu */
		{
			l2,		/* .l2 */
			0,		/* .lambda */
			s0,		/* .s0 */
			w,		/* .w */
			scratch		/* .scratch */
		}	/* .mmu_push */
	};
	struct DECORATE(cgeq_spec) problem= {
		DECORATE(minimu_grad_wrap),	/* .gradient */
		( get_delta ?
			DECORATE(minimu_precond_wrap) :
			precondition
		),				/* .precond */
		normalize,			/* .normalize */
		project,			/* .project */
		coproject,			/* .coproject */
		stepsize,			/* .guess_step */
	};

	if(get_own_scratch) {
		scratch= malloc((1+4*3)*n*sizeof(*scratch));
		if(0==scratch) return -1;
		mmpw.mmu_push.scratch= scratch;
	}
	status= DECORATE(cgeq)(
		3*n, s, scratch+n,
		rtorq2_goal, rtorq2, max_iterations,
		&problem,
		&mmpw,
		logging_hook, logging_context
	);
	*lambda= mmpw.mmu_push.lambda;
	if(get_own_scratch && scratch) free(scratch);
	return status;
} /* minimu_push() */


#ifdef TEST

/* Permalloy-ish thin film */
#include "slab_problem.c"
/* we use:
 *  struct slab;
 *  void slab_field(int n, const FLOAt s[], FLOAt h[], void *context);
 *  int slab_init_matrix(struct slab *data);
 *  int slab_gnuplot_vectordump(FILE *out, FLOAt s[], struct slab *data);
 *  void slab_release_matrix(struct slab *data);
 */

static int smoke_test(int argc, char *argv[])
{

	unsigned long slabseed= 4;
	int bad= -1;
	struct slab data= {
		.nx= 20, .ny= 20,
		.dx= 10e-7, .dy= 10e-7,	// cm
		.A= 1e-6,		// ergs/cm
		.Ku= 10.*800/2,		// ergs/cc, for Hk= 10 Oe
		.Ms= 800,		// emu/cc
		.t= 10e-7,		// cm
		.Matrix= 0,		// NULL
	};

	int ns= data.nx * data.ny;	// number of spins
	int n3= ns*3;			// number of components
	FLOAt *s= 0;			// bucket of numbers

	int i, status;
	FLOAt r2, r2goal;

	FLOAt *s0;
	FLOAt *scratch;
	FLOAt *w;

	FILE *magfile= NULL;
	char *magname= 0;
	FILE *cglog= NULL;
	char *cglogname= 0;
	FILE *gpfile= NULL;
	FILE *cg_gpfile= NULL;
	char *lambdaname= 0;
	FILE *lambdalog= NULL;
	FILE *lambd_gpfile= NULL;

	FLOAt l2;
	FLOAt ell[]= {
		.01, .02, .05,
		.1,  .2,  .3, .35, .36, .37, .38, .39,
		.40, .45, .50, .55,
		.60, .61, .62, .63, .64, .65,
		.70, .75, .80,
		.81, .82, .83, .84, .85,
		.90, .95, 1.00,
	};
	const int n_ell= sizeof(ell)/sizeof(ell[0]);
	FLOAt lambda;
	int k, l2_status, graphno;

	if(0 == verbosity) return 0;

	if(argc>1 && 0==strcmp(argv[0], "--slabseed")) {
		sscanf(argv[1], "%lu", &slabseed);
	}

	printf(
		"#\n"
		"# %s: the \"smoke test\" still requires manual inspection.\n"
		"# Check the log files listed below.\n"
		"#\n"
		, cmd
	);

	/* s[3*ns]
	 * s0[3*ns]
	 * scratch[13*ns]
	 * w[ns]
	 */
	s= malloc((3+3+13+1)*ns*sizeof(*s));
	if(0==s) return -1;
	s0= s+n3;
	scratch= s0+n3;
	w= scratch + 13*ns;

	if(0 != slab_init_matrix(&data)) {
		puts("smoke_test: initialization failure");
		bad++;
		goto out;
	}

	if(verbosity) {
		magfile= openlogfile("slab", &magname);
		cglog= openlogfile("slab-cg", &cglogname);
		gpfile= openlogfile("slab-1-mag-gp", 0);
		cg_gpfile= openlogfile("slab-2-cg-gp", 0);
		lambdalog= openlogfile("slab-lambda", &lambdaname);
		lambd_gpfile= openlogfile("slab-0-lambda-gp", 0);
		if(gpfile) fprintf(
			gpfile,
			"unset border\n"
			"unset xtics\n"
			"unset ytics\n"
			"unset key\n"
			"set size ratio %e\n"
			, 1.0*data.ny / data.nx
		);
		if(gpfile) fprintf(
			gpfile,
			"set xrange [%e : %e]\n"
			"set yrange [%e : %e]\n"
			, -1.0, 0.0+data.nx, -1.0, 0.0+data.ny
		);
		if(cg_gpfile) fprintf(
			cg_gpfile,
			"set logscale y\n"
			"set style data linespoints\n"
			"set title 'Setting the scale, forgive the flash'\n"
			"set xlabel 'iteration number\n"
			"unset key\n"
			"set ylabel 'RMS residual'\n"
			"set xrange [ *:* ] writeback\n"
			"set yrange [ *:* ] writeback\n"
			"plot '%s' using 1:(sqrt($2/%f))\n"
			"set xrange restore\n"
			"set yrange restore\n\n"
			, cglogname, (double) ns
		);
		if(lambdalog) fputs(
			"# Lagrange multiplier vs. push radius\n"
			"#\n"
			, lambdalog
		);
		if(lambd_gpfile) fprintf(
			lambd_gpfile,
			"set style data points\n"
			"set pointsize 1.5\n"
			"set zeroaxis x\n"
			"set title 'Lagrange multiplier vs. push radius'\n"
			"unset key\n"
			"plot '%s'\n"
			"pause -1\n"
			"replot\n"
			"pause -1\n"
			, lambdaname
		);
	}	/* if(verbosity) */

	for(i= 0; i<ns; i++) {
		s0[0+3*i]= 1; s0[1+3*i]= .5; s0[2+3*i]= 0;
	}
	DECORATE(minimu_normalize)(ns*3, s0, 0);

	r2goal= ns*1e-2;		// (.1 Oe)^2
	if(verbosity && cglog) fputs("# initial state\n", cglog);
	status= DECORATE(minimu)(
		ns, s0, scratch, r2goal, &r2, 500,
		slab_field, 0,
		&data,
		cgeq_simple_log_hook, cglog
	);
	printf("minimu returns %#x\n", status);
	if(verbosity) {
		if(magfile) slab_gnuplot_vectordump(magfile, s0, &data);
		if(gpfile) fprintf(gpfile,
			"set title 'reference, r2= %e  status= %#x",
			r2, status
		);
		if(gpfile) fprintf(gpfile,
			"\nplot '%s' index 0 with vectors lt 3\n"
			"pause -1\n"
			, magname
		);
		if(cg_gpfile) fprintf(cg_gpfile,
			"\nset title 'L= 0'\n"
			"plot '%s' index 0 using 1:(sqrt($2/%f))\n"
			"pause -1\n"
			, cglogname, (double) ns
		);
		if(cglog) fputs("\n\n", cglog);
	}

	for(i= 0; i<ns; i++) w[i]= 1;
	badseed(slabseed);
	for(i= 0; i<n3; i++) s[i]= s0[i] + ell[0]*(badrandom()-0.5);

	graphno= 1;
	for(k= 0; k< n_ell; k++) {
		do {
			l2_status= DECORATE(
				minimu_push_make_feasible
			)(ns, s, s0, w, ell[k]*ell[k], &l2);
			if(verbosity && cglog) fprintf(cglog,
				"# attempting L= %11.8f\n", sqrt(l2)
			);
			status= DECORATE(minimu_push)(
				ns, s, scratch,
				r2goal, &r2, 400,
				slab_field, 0,
				s0, w, l2, &lambda,
				&data,
				cgeq_simple_log_hook, cglog
			);
			printf(
				"L= %f, make_feasible= % 2d, "
				"minimu_push= %#x, lambda= %g\n",
				sqrt(l2), l2_status, status, lambda
			);
			if(lambdalog) fprintf(
				lambdalog,
				"%11.8f  %g       %#x\n",
				sqrt(l2), lambda, status
			);

			if(verbosity) {
				if(magfile) slab_gnuplot_vectordump(
					magfile, s, &data
				);
				if(gpfile) fprintf(gpfile,
					"set title \\\n"
					"'L= %f  lambda= %g   "
					"r2= %e status= %#x'\n"
					, sqrt(l2), lambda, r2, status
				);
				if(gpfile) fprintf(gpfile,
					"\nplot \\\n"
					"\t'%s' index  0 "
					"with vectors lt 6, \\\n"
					"\t'%s' index %d with vectors lt 3\n"
					"pause -1\n"
					, magname, magname, graphno
				);
				if(cg_gpfile) fprintf(cg_gpfile,
					"set title 'L= %f'\n"
					"plot '%s' index %d "
					"using 1:(sqrt($2/%f))\n"
					"pause -1\n"
					, sqrt(l2), cglogname, graphno,
					(double) ns
				);
				if(cglog) fputs("\n\n", cglog);
				graphno++;
			}
		} while(l2_status < 0);
	}

	out:
	if(lambd_gpfile) fclose(lambd_gpfile);
	if(lambdaname) free(lambdaname);
	if(gpfile) {
		fputs("pause -1\n", gpfile);
		fclose(gpfile);
	}
	if(cg_gpfile) {
		fputs("pause -1\n", cg_gpfile);
		fclose(cg_gpfile);
	}
	if(cglog) fclose(cglog);
	if(magfile) fclose(magfile);
	if(magname) free(magname);
	slab_release_matrix(&data);
	if(s) free(s);
	return bad;
} /* smoke_test() */



int (*tests[])(int argc, char *argv[])= {
	test_normalize,
	test_project,
	test_coproject,
	smoke_test,
	0
};
#endif /* TEST */
