/* $Id: minimu_push.h,v 1.5.2.1 2007/04/08 23:32:01 pa Exp $
 * Energy minimization, constrained micromagnetics.
 * Read the file "minimu.h" first.
 *
 * Copyright (C) 2007  Seagate Technology
 * See terms at the end of this file.
 */
#ifndef MINIMU_PUSH_H
#define MINIMU_PUSH_H

#include "minimu.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * We minimize the micromagnetic energy of a spin system s[],
 * subject to the additional constraint that the Euclidean distance
 * between s[] and a reference state s0[] be equal to a given value.
 */

extern int minimu_push(
	/* micromagnetic state */
	int n,			/* number of spins */
	double s[],		/* s[3*n], guess, replaced */
	double scratch[],	/* scratch[13*n], or NULL */
	/* exit criteria */
	double rtorq2_goal,	/* exit threshold */
	double *rtorq2,		/* residuals actually reached */
	int max_iterations,	/* limit not to exceed */
	/* callbacks */
	minimu_field_proc get_h,	/* field computation */
	minimu_precond    get_delta,	/* optional preconditioner */
	/* constraint specification, see below */
	const double s0[],	/* s0[3*n], reference state */
	const double w[],	/* w[n], weights */
	double l2,		/* push value away from s0[] */
	double *lambda,		/* return, Lagrange multiplier */
	/* miscellaneous */
	void *context,		/* forwarded to callbacks */
	/*
	* Next two are normally NULL, see the bottom of "cgeq.h" for usage.
	*/
	int (*logging_hook)(int opcode, void *logging_context, ...),
	void *logging_context
);

/* The constraint is
 *     \sum_{i=1}^{nspins} w[i]*||s[i]-s0[i]||^2 == l2
 * 
 * The weights w[] are internally normalized to unit sum, so
 * the allowable values for l2 are 0 < l2 < 4.
 * Zero would mean s == s0 and 4 would mean s == -s0, leaving
 * no degrees of freedom available, so l2 should be strictly
 * between 0 and 4.
 */

/* Since the normalization is nontrivial, we export a function to do
 * it.  This is NOT suitable as a callback for cgeq().
 * Returns zero on success, negative on error.
 *
 * The underlying normalization operation has a limited range, so 
 * we may fail to achieve the requested l2.  If so, we return
 * an error and we store the actual squared distance in the space
 * pointed to by the pointer l2_actual (if not NULL).  We store
 * a negative number in case of fatal error.
 */
extern int minimu_push_make_feasible(
	int n,			/* number of spins */
	double s[],		/* s[3*n], state to modify */
	const double s0[],	/* s0[3*n], reference state */
	const double w[],	/* w[n], weights */
	double l2,		/* desired squared distance */
	double *l2_actual	/* what we got, NULL ok */
);


/* Single precision */
extern int minimu_push_f(
	int n, float s[], float scratch[],
	float rtorq2_goal, float *rtorq2, int max_iterations,
	minimu_field_proc_f get_h,
	minimu_precond_f    get_delta,
	const float s0[], const float w[], float l2, float *lambda,
	void *context,
	int (*logging_hook)(int opcode, void *logging_context, ...),
	void *logging_context
);
extern int minimu_push_make_feasible_f(
	int n, float s[], const float s0[], const float w[],
	float l2, float *l2_actual
);

#ifdef __cplusplus
}
#endif

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#endif /* MINIMU_PUSH_H */
