/* $Id: slab_problem.c,v 1.8.2.1 2007/04/08 23:32:01 pa Exp $
 * Test problem for micromagnetics.
 * This file is not self-contained and will only work if #included
 * from an "#ifdef TEST" block in the micromagnetic codes.
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */
/*
 * Rectangular permalloy-ish slab, crudely discretized as a 2D array
 * of zero-size spins.  We need some "context".
 *
 * Synopsis:
 */
struct slab;
static void slab_field(int n, const FLOAt s[], FLOAt h[], void *context);
static int slab_init_matrix(struct slab *data);
static int slab_init_state(FLOAt s[], struct slab *data);
static int slab_gnuplot_vectordump(FILE *out, FLOAt s[], struct slab *data);
static void slab_release_matrix(struct slab *data);


struct slab {
	int nx, ny;	// grid size
	FLOAt dx, dy;	// grid pitch, cm
	FLOAt A;	// exch. stiffness, ergs/cm
	FLOAt Ku;	// anisotropy, along x, ergs/cm^3
	FLOAt Ms;	// emu/cm^3
	FLOAt t;	// thickness, cm
	FLOAt **Matrix;	// Matrix[3*nx][3*ny]
};

void slab_release_matrix(struct slab *data)
{
	if(data->Matrix) {
		free(data->Matrix[0]); free(data->Matrix);
	}
}

static void slab_field(
	int n, const FLOAt s[], FLOAt h[], void *context
)
{
	struct slab *data= context;
	int i,j, n3= 3*n;
	FLOAt * restrict * restrict M= data->Matrix;

	for(i= 0; i<n3; i++) h[i]= 0;

	for(i= 0; i<n3; i++) for(j= 0; j<n3; j++) {
		h[i]+= M[i][j]*s[j];
	}
}

static int slab_init_matrix(struct slab *data)
{
	int nspins= data->nx * data->ny;
	FLOAt **rows, *buf;
	int i,j, n3= 3*nspins, nn9= n3*n3;

	data->Matrix= 0;
	/* -- FIXME -- assumes 0x0 == floating point zero */
	buf= calloc(nn9, sizeof(*buf));
	if(0== buf) return -1;
	rows= malloc(n3*sizeof(*rows));
	if(0==rows) {
		free(buf);
		return -1;
	}
	for(i= 0; i<n3; i++) {
		rows[i]= buf;
		buf+= n3;
	}
	data->Matrix= rows;
	buf= rows[0];

	/* anisotropy */
	{
		FLOAt Hk= 2*data->Ku / data->Ms;
		for(i= 0; i< n3; i+= 3) rows[i][i]+= Hk;
	}

	/* Exchange.  Spins are stored in s[] in row-major order. */
	{
		FLOAt hex= data->A/(data->Ms * data->dx * data->dx);
		FLOAt hey= data->A/(data->Ms * data->dy * data->dy);
		// FLOAt tmp;
		int ntx= data->nx-1, nty= data->ny-1;	// tiles, not spins
		int ix,iy;
		int i00,i10,i01,i11;
		int k;
		for(iy= 0; iy<nty; iy++) {
			for(ix= 0; ix<ntx; ix++) {
				i00= 3*(ix + (data->nx)*iy);
				i10= i00+3;
				i01= i00 + 3*(data->nx);
				i11= i01 + 3;
				for(k= 0; k<3; k++) {
					rows[k+i10][k+i10] -= hex;
					rows[k+i10][k+i00] += hex;
					rows[k+i00][k+i10] += hex;
					rows[k+i00][k+i00] -= hex;
					// tmp= hex*(s[k+i10]-s[k+i00]);
					// h[k+i10]-= tmp; h[k+i00]+= tmp;

					rows[k+i11][k+i11] -= hex;
					rows[k+i11][k+i01] += hex;
					rows[k+i01][k+i11] += hex;
					rows[k+i01][k+i01] -= hex;
					// tmp= hex*(s[k+i11]-s[k+i01]);
					// h[k+i11]-= tmp; h[k+i01]+= tmp;

					rows[k+i01][k+i01] -= hey;
					rows[k+i01][k+i00] += hey;
					rows[k+i00][k+i01] += hey;
					rows[k+i00][k+i00] -= hey;
					// tmp= hey*(s[k+i01]-s[k+i00]);
					// h[k+i01]-= tmp; h[k+i00]+= tmp;

					rows[k+i11][k+i11] -= hey;
					rows[k+i11][k+i10] += hey;
					rows[k+i10][k+i11] += hey;
					rows[k+i10][k+i10] -= hey;
					// tmp= hey*(s[k+i11]-s[k+i10]);
					// h[k+i11]-= tmp; h[k+i10]+= tmp;
				}
			}
		}
	}

	/* magnetostatics */
	{
		int nx= data->nx, ny= data->ny, n= nx*ny;
		int is, isx,isy, js, jsx,jsy;
		FLOAt dx= data->dx, dy= data->dy;
		FLOAt xij, yij;
		FLOAt r2ij, rij;		// inverses, actually
		FLOAt scale= data->Ms * dx * dy * data->t;
		FLOAt tmp;  // FLOAt tmp2;
		for(is= 0; is<n; is++) {
			i= 3*is; isy= is/nx; isx= is-nx*isy;
			for(js= 0; js<is; js++) {
				j= 3*js; jsy= js/nx; jsx= js-nx*jsy;

				xij= dx*(jsx-isx); yij= dy*(jsy-isy);
				r2ij= 1/(xij*xij + yij*yij);
				rij= SQRT(r2ij);


				// -\delta_{ij} / r^3
				tmp= -scale*rij*r2ij;
				rows[0+i][0+j]+= tmp;
				rows[1+i][1+j]+= tmp;
				rows[2+i][2+j]+= tmp;
				rows[0+j][0+i]+= tmp;
				rows[1+j][1+i]+= tmp;
				rows[2+j][2+i]+= tmp;
				// h[0+i]+= tmp*s[0+j];
				// h[1+i]+= tmp*s[1+j];
				// h[2+i]+= tmp*s[2+j];
				// h[0+j]+= tmp*s[0+i];
				// h[1+j]+= tmp*s[1+i];
				// h[2+j]+= tmp*s[2+i];

				// 3 ri rj / r^5
				tmp*= -3*r2ij;
				rows[0+i][0+j] += tmp*xij*xij;
				rows[0+i][1+j] += tmp*xij*yij;
				rows[1+i][0+j] += tmp*xij*yij;
				rows[1+i][1+j] += tmp*yij*yij;
				// tmp2= tmp*(xij*s[0+j] + yij*s[1+j]);
				// h[0+i]+= xij*tmp2;
				// h[1+i]+= yij*tmp2;
				rows[0+j][0+i] += tmp*xij*xij;
				rows[0+j][1+i] += tmp*xij*yij;
				rows[1+j][0+i] += tmp*xij*yij;
				rows[1+j][1+i] += tmp*yij*yij;
				// tmp2= tmp*(xij*s[0+i] + yij*s[1+i]);
				// h[0+j]+= xij*tmp2;
				// h[1+j]+= yij*tmp2;
			}
		}
	}
	return 0;
}	/* slab_init_matrix() */

int slab_init_state(FLOAt s[], struct slab *data)
{
	/* initialize bottom half magnetized right, top magnetized left */
	int i, ix, iy;
	for(i= 0, iy= 0; iy<(data->ny/2); iy++) {
		for(ix= 0; ix<data->nx; ix++, i+= 3) {
			s[i]= 1; s[i+1]= s[i+2]= 0;
		}
	}
	for(; iy<data->ny; iy++) {
		for(ix= 0; ix<data->nx; ix++, i+= 3) {
			s[i]= -1; s[i+1]= s[i+2]= 0;
		}
	}
	return DECORATE(minimu_normalize)(3*(data->nx)*(data->ny), s, 0); 
}

int slab_gnuplot_vectordump(FILE *out, FLOAt s[], struct slab *data)
{
	int i, ix,iy;
	for(i= 0, iy= 0; iy < (data->ny); iy++) {
		for(ix= 0; ix < (data->nx); ix++, i+= 3) {
			fprintf(out,
				"% f  % f    % f  % f\n",
				ix-.4*s[0+i], iy-.4*s[1+i],
				.8*s[0+i], .8*s[1+i]
			);
		}
	}
	fputs("\n\n", out);
	return 0;
}
