/* $Id: test-rig.c,v 1.19.2.1 2007/04/08 23:32:01 pa Exp $
 * Boilerplate for unit tests.
 * See comments at the bottom for general philosophy.
 *
 * Copyright (C) 2007  Seagate Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * The author can be reached by electronic mail at Pierre.Asselin@seagate.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "config.h"

extern int (*tests[])(int argc, char *argv[]);

static int verbosity= 0;
static char *cmd;
static double epsilon;
static float epsilon_f;

int parse_long(const char *option);	// see below
int main(int argc, char *argv[])
{
	int i, ntests;
	int bad= 0;

	/* figure out machine epsilon */
	{
		volatile double vd;
		volatile float vf;
		epsilon= 1; epsilon_f= 1;
		for(;;) {
			epsilon*= 0.5;
			vd= 1+epsilon; vd-= 1;
			if(0 == vd) break;
		}
		for(;;) {
			epsilon_f*= 0.5F;
			vf= 1+epsilon_f; vf-= 1;
			if(0 == vf) break;
		}
	}

	/* count the tests */
	for(ntests= 0; tests[ntests]; ntests++) /* nada */;

	/* Save argv[0] */
	cmd= argv[0];

	/* Walk through command line, stop at first unrecognized option */
	while(++argv, --argc) {
		if('-' != argv[0][0]) goto done;
		switch(argv[0][1]) {
		case '-':
			if(0 != parse_long(argv[0]+2)) goto done;
		break;
		case 'v':
			verbosity++;
		break;
		default:
			goto done;
		}
	}
	done:

	/* run through the tests */
	for(i= 0; i<ntests; i++) bad|= tests[i](argc, argv);

	return !!bad;
}

/* this stolen from the rand(3) manpage */
static unsigned long badstate= 1;
double badrandom(void)
{
	unsigned long out;
	badstate= badstate * 1103515245 + 12345;
	out= (unsigned)(badstate/65536) % 32768;
	return out/32768.0;
}
void badseed(unsigned long i) {badstate= i;}


int parse_long(const char *option)
{
	unsigned long ultmp;
	int itmp;

	if(1 == sscanf(option, "seed=%lu", &ultmp)) {
		badseed(ultmp);
		return 0;
	} else if(1 == sscanf(option, "verbosity=%d", &itmp)) {
		verbosity= itmp;
		return 0;
	}

	return -1;
}


#include <assert.h>
static char *mangled_cmd= 0;
FILE *openlogfile(const char suffix[], char **return_filename)
{
	char *filename;
	FILE *stream;
	int l_myname;
	int l_suffix= strlen(suffix);
	int l_filename;

	if(0==mangled_cmd) {
#ifndef HAVE_WINDOWS_H
		char *c= strrchr(cmd, '/');
#else
		char *c= strrchr(cmd, '\\');
#endif
		if(c) c++; else c= cmd;
		mangled_cmd= malloc(1+strlen(c));
		strcpy(mangled_cmd, c);
		c= strrchr(mangled_cmd, '.');
		if(c) *c= 0;
		c= strchr(mangled_cmd, '_');
		if(c) *c= '-';
	}
	l_myname= strlen(mangled_cmd);

	/* concatenate mangled_cmd, "-", suffix, ".out" */
	l_filename= l_myname + 1 + l_suffix + 4;
	filename= malloc(l_filename+1);
	if(0==filename) return 0;
	strncpy(filename, mangled_cmd, l_myname);
	filename[l_myname]= '-';
	strcpy(filename+l_myname+1, suffix);
	strcpy(filename+l_myname+1+l_suffix, ".out");

	assert(strlen(filename) == l_filename);
	errno= 0;
	stream= fopen(filename, "w");
	if(0 == stream) perror(filename);
	if(verbosity) printf("(see %s)\n", filename);

	if(return_filename) *return_filename= filename;
	else free(filename);
	setvbuf(stream, (char *)0, _IOLBF, 0);
	return stream;
}


/* Okay, here's how this works.
 *
 * All code should be accompanied by test functions (unit tests) that
 * verify its operation.  The test code does not belong in the final
 * product, so it is placed in conditional compilation blocks,
 * something like
 *
 *   // good code ...
 *   #ifdef TEST
 *   // testing code goes here
 *   #endif // TEST
 *
 * The idea is that the compilation command
 *   cc -c file.c
 * produces a normal object "file.o", while
 *   cc -DTEST file.c
 * produces a self-contained executable "a.out" that runs a torture test
 * on the normal code.
 *
 * Unit tests quickly turn into repetitive code, so we look for
 * typical boilerplate like parsing the command line or keeping track
 * of cumulative errors and yank them into a separate file, this file.
 *
 * To work with this file, the unit tests must be functions similar to
 * a main program, i.e.  "int sometest(int argc, char *argv[])".
 * Return values are zero for pass, anything else for fail.
 *
 * SOMEBODY MUST DEFINE A NULL-TERMINATED ARRAY OF POINTERS TO THE
 * TESTS, CALLED "tests".  We DECLARE tests[] above, but somebody else
 * must DEFINE it.  Usually the definition will be at the bottom of an
 * "#ifdef TEST" conditional block in the code under test, something like
 *
 *    static int (*tests[])(int argc, char *argv[])= {
 *        my_first_test,
 *        my_second_test,
 *        test_etc,
 *        etc,
 *        0	// terminating NULL.  IMPORTANT !!
 *    };
 *
 * where all the array entries are functions defined in some other
 * "#ifdef TEST" block, or maybe even in the same one.
 *
 *    static int my_first_test(int argc, char *argv[]) {
 *        int status= 0
 *        // try to break things
 *        return status;
 *    }
 *
 * That way, we can write the main program above without knowing
 * anything about the tests or even how many there are.
 *
 * A test that passes should be silent by default.  A failed test should
 * carp to stdout (or stderr ?) in addition to returning a nonzero value
 * so the maintainer can see that there is a problem and fix the code
 * under test.
 *
 * It would be nice to have crystal-clear error messages, but in practice
 * they're going to be cryptic.  Whoever discovers a failure may need help
 * from the primary maintainer.  (Also, be advised that the fault is often
 * in the test, not the code under test !)
 *
 * The main program above scans the command line (argc, argv) for useful
 * options and sets global variables accordingly.  Tests can tailor
 * their behavior based on said variables, for example they can decide
 * to quack a little even if they pass, based on the "verbosity"
 * variable.
 *
 * The openlogfile() returns a stream ready for writing, for tests
 * that want to spew large amounts of output on the side.  I
 * centralize it here because:
 *   1)  it's boilerplate
 *   2)  it enforces a naming convention (cmd-suffix.out) that
 *       Makefiles can use to clean up those files.
 *
 * Global variables are convenient and bad.  We use them even though
 * they're bad because these are *unit* tests that will never grow
 * into big software engineering projects.  The global variables are
 * compiled conditionally so they do not pollute the payload code.
 *
 * The main program strips the command name and options it knows about
 * from (argc, argv) and passes the rest to the tests.
 * Most tests will ignore their arguments.  Occasionally
 * there is a benefit to fine-tuning at run-time, in which case the
 * function implementing the test has to parse argv[] like a main
 * program would (except that argv[0] is the first option, not the
 * command name).  Thus, we can have generic options and test-specific
 * options.  If this gets out of hand, the individual tests have to
 * recognize a "-help" or "--help" option and describe the choices
 * they understand.
 *
 * In summary, a source file with built-in tests will look like this:
 *
 *     // good code goes here...
 *
 *     #ifdef TEST
 *     #include "test-rig.c"
 *     // buch of unit test definitions
 *     #endif // TEST
*
 *     // more good code...
 *
 *     #ifdef TEST
 *     // more unit tests ...
 *     #endif // TEST
 *
 *     // and so forth
 *
 *     #ifdef TEST
 *     // final bunch of unit tests ...
 *
 *     // ... and the array of pointers to them all
 *     static int (*tests[])(int argc, char *argv[])= {
 *         the_first_test, the_second_test, // ...
 *         0				    // do NOT leave out the NULL
 *     };
 *     #endif // TEST
 */
